# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import cython
import numpy as np

from .general import AbstractDetector


class FKMDetector(AbstractDetector):
    """Rainflow detector as described in FKM non linear.

    The algorithm has been published by Clormann & Seeger 1985 and has
    been cited heavily since.

    .. jupyter-execute::

        from pylife.stress.timesignal import TimeSignalGenerator
        import pylife.stress.rainflow as RF

        ts = TimeSignalGenerator(10, {
            'number': 50,
            'amplitude_median': 1.0, 'amplitude_std_dev': 0.5,
            'frequency_median': 4, 'frequency_std_dev': 3,
            'offset_median': 0, 'offset_std_dev': 0.4}, None, None).query(10000)

        rfc = RF.FKMDetector(recorder=RF.LoopValueRecorder())
        rfc.process(ts)

        rfc.recorder.collective

    Alternatively you can ask the recorder for a histogram matrix:

    .. jupyter-execute::

        rfc.recorder.histogram(bins=16)

    Note
    ----
    This detector **does not** report the loop index.
    """

    def __init__(self, recorder):
        """Instantiate a FKMDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        super().__init__(recorder)
        self._ir = 1
        self._residuals = []
        self._max_turn = 0.0

    @cython.locals(
        turns=cython.double[:],
        iz=cython.int, ir=cython.int,
        last0=cython.double, last1=cython.double,
        loop_assumed=cython.int,
        max_turn=cython.double)
    def process(self, samples):
        """Process a sample chunk.

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : FKMDetector
            The ``self`` object so that processing can be chained
        """
        ir = self._ir
        max_turn = self._max_turn
        turns_index, turns = self._new_turns(samples)

        from_vals = []
        to_vals = []

        for current in turns:
            loop_assumed = True
            while loop_assumed:
                iz = len(self._residuals)
                if iz < ir:
                    break
                loop_assumed = False
                if iz > ir:
                    last0 = self._residuals[-1]
                    last1 = self._residuals[-2]
                    if np.abs(current-last0) >= np.abs(last0-last1):
                        from_vals.append(last1)
                        to_vals.append(last0)
                        self._residuals.pop()
                        self._residuals.pop()
                        if np.abs(last0) < max_turn and np.abs(last1) < max_turn:
                            loop_assumed = True
                    continue
                if np.abs(current) > max_turn:
                    ir += 1
            max_turn = max(np.abs(current), max_turn)
            self._residuals.append(current)

            self._ir = ir
            self._max_turn = max_turn

        self._recorder.record_values(from_vals, to_vals)

        return self



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Meanstress routines
===================

Mean stress transformation methods
----------------------------------

* FKM Goodman
* Five Segment Correction

'''

__author__ = "Johannes Mueller, Lena Rapp"
__maintainer__ = "Johannes Mueller"

from collections.abc import Iterable
import operator as op

import numpy as np
import pandas as pd

from pylife import PylifeSignal, Broadcaster
import pylife.stress.collective as CL


@pd.api.extensions.register_series_accessor('haigh_diagram')
class HaighDiagram(PylifeSignal):
    """Model for a Haigh diagram in order to perform meanstress transformations.

    A Haigh diagram a set of meanstress sensitivity slopes $M$ that is changing
    with the R-values.  The values of the ```pd.Series`` represents that slopes
    $M$ and the `pd.IntervalIndex` represents the R-ranges.
    """

    @classmethod
    def from_dict(cls, segments_dict):
        """Create a Haigh diagram from a dict.

        Parameters
        ----------
        segments_dict : dict
            dict resolving the R-value intervals to the meanstress slope

        Example
        -------
        >>> hd = MST.HaighDiagram.from_dict({
        >>>    (1.0, np.inf): 0.0,
        >>>    (-np.inf, 0.0): 0.5,
        >>>    (0.0, 1.0): 0.167
        >>> })

        sets up a FKM Goodman like Haigh diagram.
        """
        vals = np.array(list(segments_dict.values()))
        idx = pd.IntervalIndex.from_tuples(list(segments_dict.keys()), name='R')
        return HaighDiagram(pd.Series(vals, index=idx))

    @classmethod
    def fkm_goodman(cls, haigh_fkm_goodman):
        """Create a Haigh diagram according to FKM Goodman.

        Parameters
        ----------
        haigh_fkm_goodman : pd.Series or pd.DataFrame
            a series containing one or a dataframe containing multiple values for
            `M` and optionally `M2`.

        The Haigh diagram according to FKM Goodman comes with the slope ``M``
        which is valid between ``R==-inf`` and ``R==0``.  Beyond ``R==0`` the slope
        is ``M2` if ``M2`` is given or ``M/3`` if not.
        """
        if 'M2' not in haigh_fkm_goodman:
            haigh_fkm_goodman['M2'] = haigh_fkm_goodman['M'] / 3.0

        M = haigh_fkm_goodman.M
        M2 = haigh_fkm_goodman.M2
        interval_index = pd.IntervalIndex.from_tuples([(1.0, np.inf), (-np.inf, 0.0), (0.0, 1.)], name='R')

        if isinstance(haigh_fkm_goodman, pd.Series):
            haigh_index = interval_index
            dummy_index = pd.Index([0, 1, 2], name='R')
        else:
            haigh_frame, _ = Broadcaster(haigh_fkm_goodman.index.to_frame()).broadcast(interval_index.to_frame())
            haigh_index = haigh_frame.index
            dummy_index = pd.Index([0, 1, 2] * len(haigh_fkm_goodman), name='R')

        haigh = pd.Series(0.0, index=dummy_index)

        R_index = haigh.index.get_level_values('R')

        haigh.iloc[R_index.get_indexer_for([1])] = M
        haigh.iloc[R_index.get_indexer_for([2])] = M2

        haigh.index = haigh_index
        return cls(haigh)

    @classmethod
    def five_segment(cls, five_segment_haigh_diagram):
        """Create a five segment slope Haigh diagram.

        Parameters
        ----------
        five_segment_haigh_diagram : :class:`pandas.Series` or :class:`pandas.DataFrame`
            The five segment meanstress slope data.

        Notes
        -----
        ``five_segment_hagih_diagram`` has to provide the following keys:
            * ``M0``: the mean stress sensitivity between ``R==-inf`` and ``R==0``
            * ``M1``: the mean stress sensitivity between ``R==0`` and ``R==R12``
            * ``M2``: the mean stress sensitivity betwenn ``R==R12`` and ``R==R23``
            * ``M3``: the mean stress sensitivity between ``R==R23`` and ``R==1``
            * ``M4``: the mean stress sensitivity beyond ``R==1``
            * ``R12``: R-value between ``M1`` and ``M2``
            * ``R23``: R-value between ``M2`` and ``M3``
        """
        was_series = isinstance(five_segment_haigh_diagram, pd.Series)

        if was_series:
            five_segment_haigh_diagram = pd.DataFrame(five_segment_haigh_diagram).T

        index_names = five_segment_haigh_diagram.index.names + ['R']

        def make_index(h):
            orig_index = [h.name] if not isinstance(h.name, Iterable) else list(h.name)

            return pd.MultiIndex.from_tuples([
                tuple(orig_index + [pd.Interval(1.0, np.inf)]),
                tuple(orig_index + [pd.Interval(-np.inf, 0.)]),
                tuple(orig_index + [pd.Interval(0., h.R12)]),
                tuple(orig_index + [pd.Interval(h.R12, h.R23)]),
                tuple(orig_index + [pd.Interval(h.R23, 1.)]),
            ], names=index_names).to_frame()

        haigh_index = pd.concat(list(five_segment_haigh_diagram.apply(make_index, axis=1))).index

        haigh = pd.Series(0.0, index=haigh_index)

        R_index = haigh.index.get_level_values('R')
        h, _ = Broadcaster(haigh).broadcast(five_segment_haigh_diagram)

        M4_locs = R_index.get_indexer_for([pd.Interval(1.0, np.inf)])
        M0_locs = R_index.get_indexer_for([pd.Interval(-np.inf, 0.)])
        M1_locs = R_index.get_indexer_for([pd.Interval(0., R12) for R12 in h.R12])
        M2_locs = R_index.get_indexer_for([pd.Interval(R12, R23) for R12, R23 in zip(h.R12, h.R23)])
        M3_locs = R_index.get_indexer_for([pd.Interval(R23, 1.) for R23 in h.R23])

        haigh.iloc[M4_locs] = h.M4.iloc[M4_locs]
        haigh.iloc[M0_locs] = h.M0.iloc[M0_locs]
        haigh.iloc[M1_locs] = h.M1.iloc[M1_locs]
        haigh.iloc[M2_locs] = h.M2.iloc[M2_locs]
        haigh.iloc[M3_locs] = h.M3.iloc[M3_locs]

        if was_series:
            haigh = haigh.xs(0)

        return cls(haigh)

    def transform(self, cycles, R_goal):
        """Transform a load collective to defined R-value.

        Parameters
        ----------
        cycles : :class:`pd.Series` accepted by class:``LoadCollective` or class:`LoadHistogram``
            The load collective

        Returns
        -------
        transformed_cycles : :class:`pd.Series`
            The transformed cycles
        """
        cycles, obj = self.broadcast(cycles, droplevel=['R'])

        transformer = _SegmentTransformer(cycles, obj, self._R_index, R_goal)

        for interval in transformer.segments_left_from_R_goal():
            interval_boundary = interval.right if interval.right < 1.0 else interval.left
            transformer.transform_cycles_in_interval(interval, interval_boundary)

        for interval in transformer.segments_right_from_R_goal():
            transformer.transform_cycles_in_interval(interval, interval.left)

        for interval in transformer.segments_containing_R_goal():
            transformer.transform_cycles_in_interval(interval, R_goal)

        transfomed_cycles = transformer.transformed_cycles
        res = pd.DataFrame({
            'range': 2. * transfomed_cycles.amplitude,
            'mean': transfomed_cycles.amplitude * ((1.+transfomed_cycles.R)/(1.-transfomed_cycles.R)).fillna(-1.0)
        }, index=cycles.index)

        return res

    def _validate(self):

        def has_gaps(idx):
            if len(idx) <= 1:
                return False
            return (pd.DataFrame({'l': idx.left[1:], 'r': idx.right[:-1]})
                    .apply(lambda r: r.l != r.r and not (r.l == -np.inf and r.r == np.inf), axis=1)
                    .any())

        self._R_index = self._find_R_index()

        if self._check_if_R_index(lambda idx: idx.is_overlapping):
            raise AttributeError("The intervals of the 'R' IntervalIndex must not overlap.")

        if self._check_if_R_index(has_gaps):
            raise AttributeError("The intervals of the 'R' IntervalIndex must not have gaps.")

    def _find_R_index(self):
        if 'R' not in self._obj.index.names:
            raise AttributeError("A Haigh Diagram needs an index level 'R'.")
        if isinstance(self._obj.index, pd.MultiIndex):
            R_index = self._obj.index.unique('R')
        else:
            R_index = self._obj.index
        if not isinstance(R_index, pd.IntervalIndex):
            raise AttributeError("The 'R' index must be an IntervalIndex.")
        return R_index

    def _check_if_R_index(self, check_func):
        if isinstance(self._obj.index, pd.IntervalIndex):
            return check_func(self._obj.index)

        all_but_R = [n or 0 for n in self._obj.index.names if n != 'R']

        # TODO: remove once we drop python 3.8 support
        include_groups = {} if pd.__version__ < "2.2" else {'include_groups': False}

        return (
            self
            ._obj.index.to_frame(index=False)
            .groupby(all_but_R)
            .apply(lambda g: check_func(g.set_index('R').index), **include_groups)
            .any()
        )


class _SegmentTransformer:

    def __init__(self, cycles, haigh, R_segments, R_goal):
        rf = cycles.load_collective
        self.transformed_cycles = pd.DataFrame({
            'amplitude': rf.amplitude,
            'R': rf.R
        }, index=cycles.index)
        self._haigh = haigh
        self._R_index = R_segments
        self._R_goal = R_goal

        self._distances = self._distance_from_R_goal()

    def segments_left_from_R_goal(self):
        return self._distances[self._distances < 0.].sort_values(ascending=True).index

    def segments_right_from_R_goal(self):
        return self._distances[self._distances > 0.].sort_values(ascending=False).index

    def segments_containing_R_goal(self):
        goal_segments = self._R_index.contains(self._R_goal)
        if not goal_segments.any():
            goal_segments = self._R_index.set_closed('left').contains(self._R_goal)

        return self._R_index[goal_segments]

    def _distance_from_R_goal(self):
        def fake_meanstress(R):
            return (1.+R)/(1.-R)

        meanstress = fake_meanstress(self._R_index.mid).fillna(-1.0)
        meanstress_goal = -1.0 if self._R_goal == -np.inf else fake_meanstress(self._R_goal)

        return pd.Series(meanstress.values - meanstress_goal, index=self._R_index)

    def transform_cycles_in_interval(self, interval, R_goal):
        def push_over_flipping_point(R):
            if R == -np.inf and R_goal > 1.0:
                return np.inf
            if R == np.inf and R_goal < 1.0:
                return -np.inf
            return R

        def cycles_in_current_interval():
            R = self.transformed_cycles.R.apply(push_over_flipping_point)
            test_interval = pd.Interval(interval.left, interval.right, closed='both')
            return R.apply(lambda R: R in test_interval)

        def cycles_in_current_segments(in_test_interval, segments_index):
            in_segments_index = pd.Series(False, index=self.transformed_cycles.index)
            in_segments_index[segments_index] = True

            return in_test_interval & in_segments_index

        def meanstress_sensitivity_segments_of_current_interval():
            return self._haigh.xs(interval, level='R')

        def transformed_amplitude():
            rf = self.transformed_cycles.loc[to_shift]
            amp = rf.amplitude
            mean = amp * (1.+rf.R)/(1.-rf.R)
            mean[rf.R == -np.inf] = -amp[rf.R == -np.inf]
            mean[rf.R == 1.0] = -amp[rf.R == 1.0]

            if R_goal == -np.inf:
                trans_amp = (amp + M * mean) / (1. - M)
            else:
                trans_amp = (1. - R_goal) * (amp + M*mean) / (1. - R_goal + M*(1.+R_goal))

            return trans_amp.fillna(0.0)

        to_shift = cycles_in_current_interval()
        if not to_shift.any():
            return

        M = meanstress_sensitivity_segments_of_current_interval()

        to_shift = cycles_in_current_segments(to_shift, M.index)
        if not to_shift.any():
            return

        if R_goal == 1.0:
            R_goal = -np.inf

        self.transformed_cycles.loc[to_shift, 'amplitude'] = transformed_amplitude()
        self.transformed_cycles.loc[to_shift, 'R'] = R_goal


def experimental_mean_stress_sensitivity(sn_curve_R0, sn_curve_Rn1, N_c=np.inf):
    r"""Estimate the mean stress sensitivity from two `FiniteLifeCurve` objects for the same amount of cycles `N_c`.

    The formula for calculation is taken from: "Betriebsfestigkeit", Haibach, 3. Auflage 2006

    Formula (2.1-24):

    .. math::
        M_{\sigma} = {S_a}^{R=-1}(N_c) / {S_a}^{R=0}(N_c) - 1

    Alternatively the mean stress sensitivity is calculated based on both SD values
    (if N_c is not given).

    Parameters
    ----------
    sn_curve_R0: pylife.strength.sn_curve.FiniteLifeCurve
        Instance of FiniteLifeCurve for R == 0
    sn_curve_Rn1: pylife.strength.sn_curve.FiniteLifeCurve
        Instance of FiniteLifeCurve for R == -1
    N_c: float, (default=np.inf)
        Amount of cycles where the amplitudes should be compared.
        If N_c is higher than a fatigue transition point (ND) for the SN-Curves, SD is taken.
        If N_c is None, SD values are taken as stress amplitudes instead.

    Returns
    -------
    float
        Mean stress sensitivity M_sigma

    Raises
    ------
    ValueError
        if the resulting M_sigma doesn't lie in the range from 0 to 1 a ValueError is raised, as this value would
        suggest higher strength with additional loads.
    """
    S_a_R0 = sn_curve_R0.woehler.basquin_load(N_c) if N_c < sn_curve_R0.ND else sn_curve_R0.SD
    S_a_Rn1 = sn_curve_Rn1.woehler.basquin_load(N_c) if N_c < sn_curve_Rn1.ND else sn_curve_Rn1.SD
    M_sigma = S_a_Rn1 / S_a_R0 - 1
    if not 0 <= M_sigma <= 1:
        raise ValueError("M_sigma: %.2f exceeds the interval [0, 1] which is not plausible." % M_sigma)
    return M_sigma


@pd.api.extensions.register_dataframe_accessor('meanstress_transform')
class MeanstressTransformCollective(CL.LoadCollective):

    def fkm_goodman(self, ms_sens, R_goal):
        hd = HaighDiagram.fkm_goodman(ms_sens)
        res = hd.transform(self._obj, R_goal)
        return res.load_collective

    def five_segment(self, haigh, R_goal):
        hd = HaighDiagram.five_segment(haigh)
        res = hd.transform(self._obj, R_goal)
        return res.load_collective


@pd.api.extensions.register_series_accessor('meanstress_transform')
class MeanstressTransformMatrix(CL.LoadHistogram):

    def _validate(self):
        super()._validate()

        if set(self._obj.index.names).issuperset({'from', 'to'}):
            f = self._obj.index.get_level_values('from').mid
            t = self._obj.index.get_level_values('to').mid
            self._Sa = np.abs(f-t)/2.
            self._Sm = (f+t)/2.
            self._binsize_x = self._obj.index.get_level_values('from').length.min()
            self._binsize_y = self._obj.index.get_level_values('to').length.min()
            self._remaining_names = list(filter(lambda n: n not in ['from', 'to'], self._obj.index.names))
        else:
            self._Sa = self._obj.index.get_level_values('range').mid / 2.
            self._Sm = self._obj.index.get_level_values('mean').mid
            self._binsize_x = self._obj.index.get_level_values('range').length.min()
            self._binsize_y = self._obj.index.get_level_values('mean').length.min()
            self._remaining_names = list(filter(lambda n: n not in ['range', 'mean'], self._obj.index.names))

    def fkm_goodman(self, haigh, R_goal):
        ranges = HaighDiagram.fkm_goodman(haigh).transform(self._obj, R_goal)['range']
        return self._rebin_results(ranges, R_goal).load_collective

    def _rebin_results(self, ranges, R_goal):

        def resulting_intervals():
            ranges_max = ranges.max()
            binsize = np.hypot(self._binsize_x, self._binsize_y) / np.sqrt(2.)
            bincount = int(np.ceil(ranges_max / binsize))
            range_bins = np.linspace(0, ranges_max, bincount+1)
            means_bins = range_bins * (1. + R_goal) / (2.0 * (1. - R_goal))
            range_itv = pd.IntervalIndex.from_breaks(range_bins, name="range")
            means_itv = pd.IntervalIndex.from_breaks(means_bins, name="mean")
            return range_itv, means_itv

        def aggregate_on_projection(projection, itvs, ranges, obj):
            level_values = tuple(projection)
            level_names = list(projection.index)
            ranges = ranges.xs(level_values, level=level_names)
            obj = obj.xs(level_values, level=level_names)
            sums = (
                itvs
                .to_series()
                .apply(lambda iv: sum_intervals(iv, ranges, obj))
            )
            return sums

        def sum_intervals(iv, ranges, obj):
            op_left = op.ge if iv.left == 0.0 else op.gt
            return obj.iloc[op_left(ranges.values, iv.left) & op.le(ranges.values, iv.right)].sum()

        if ranges.shape[0] == 0:
            new_idx = pd.IntervalIndex(pd.interval_range(0.,  0., 0), name='range')
            return pd.Series([], index=new_idx, name='cycles', dtype=np.float64)

        range_itv_idx, means_itg_idx = resulting_intervals()

        if len(self._remaining_names) > 0:
            remaining_idx = self._obj.index.to_frame(index=False).groupby(self._remaining_names).first().index
            result = (
                remaining_idx.to_frame(index=False)
                .apply(lambda p: aggregate_on_projection(p, range_itv_idx, ranges, self._obj), axis=1)
            )
            result.index = remaining_idx
            result.columns = pd.MultiIndex.from_arrays([result.columns, means_itg_idx])
            result = result.stack(['range', 'mean']).reorder_levels(['range', 'mean'] + self._remaining_names)
        else:
            result = range_itv_idx.to_series().apply(lambda iv: sum_intervals(iv, ranges, self._obj))
            result.index = pd.MultiIndex.from_arrays([result.index, means_itg_idx])

        return result


def fkm_goodman(amplitude, meanstress, M, M2, R_goal):
    cycles = pd.DataFrame({
        'range': 2.*amplitude,
        'mean': meanstress
    })

    haigh_fkm_goodman = pd.Series({
        'M': M,
        'M2': M2
    })
    hd = HaighDiagram.fkm_goodman(haigh_fkm_goodman)

    res = hd.transform(cycles, R_goal)
    return res.load_collective.amplitude.to_numpy()


def five_segment_correction(amplitude, meanstress, M0, M1, M2, M3, M4, R12, R23, R_goal):
    ''' Performs a mean stress transformation to R_goal according to the
        Five Segment Mean Stress Correction

    :param Sa: the stress amplitude
    :param Sm: the mean stress
    :param Rgoal: the R-value to transform to
    :param M: the mean stress sensitivity between R=-inf and R=0
    :param M1: the mean stress sensitivity between R=0 and R=R12
    :param M2: the mean stress sensitivity betwenn R=R12 and R=R23
    :param M3: the mean stress sensitivity between R=R23 and R=1
    :param M4: the mean stress sensitivity beyond R=1
    :param R12: R-value between M1 and M2
    :param R23: R-value between M2 and M3

    :returns: the transformed stress range
    '''

    cycles = pd.DataFrame({
        'range': 2.*amplitude,
        'mean': meanstress
    })

    haigh_five_segment = pd.Series({
        'M0': M0,
        'M1': M1,
        'M2': M2,
        'M3': M3,
        'M4': M4,
        'R12': R12,
        'R23': R23
    })

    hd = HaighDiagram.five_segment(haigh_five_segment)
    res = hd.transform(cycles, R_goal)
    return res.load_collective.amplitude.to_numpy()



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Cedric Philip Wagner"
__maintainer__ = "Johannes Mueller"

import warnings

import pandas as pd

import pylife.strength.fatigue
import pylife.stress

warnings.warn(
    FutureWarning(
        "The module pylife.strength.helpers is deprecated and no longer under test. "
        "The functionality is now avaliable in the pylife.materiallaws.WoehlerCurve."
    )
)

class FiniteLifeBase:
    """Base class for SN curve calculations - either in logarithmic or regular scale"""

    def __init__(self, k_1, SD_50, ND_50):
        warnings.warn(DeprecationWarning("FiniteLifeBase and derived classes are deperecated. "
                                         "Use WoehlerCurve and Fatigue accessors instead."))
        self._wc = pd.Series({
            'k_1': k_1,
            'SD': SD_50,
            'ND': ND_50
        })

    @property
    def k_1(self):
        return self._wc.k_1


class FiniteLifeLine(FiniteLifeBase):
    """Sample points on the finite life line - either N or S (LOGARITHMIC SCALE)

    The formula for calculation is taken from
    "Betriebsfestigkeit", Haibach, 3. Auflage 2006

    Notes
    -----
    In contrast to the case

    Parameters
    ----------
    k : float
        slope of the SN-curve
    SD_50 : float
        lower stress limit in the finite life region
    ND_50 : float
        number of cycles at stress SD_50
    """

    def __init__(self, k, SD_50, ND_50):
        super().__init__(k, SD_50, ND_50)


class FiniteLifeCurve(FiniteLifeBase):
    """Sample points on the finite life curve - either N or S (NOT logarithmic scale)

    The formula for calculation is taken from
    "Betriebsfestigkeit", Haibach, 3. Auflage 2006

    **Consider:** load collective and life curve have to be consistent:

        * range vs range
        * amplitude vs amplitude

    Parameters
    ----------
    k : float
        slope of the SN-curve
    SD_50 : float
        lower stress limit in the finite life region
    ND_50 : float
        number of cycles at stress SD_50
    """
    def __init__(self, k_1, SD_50, ND_50):
        super().__init__(k_1, SD_50, ND_50)

    def calc_S(self, N, ignore_limits=True):
        """Calculate stress S for a given number of cycles N

        Parameters
        ----------
        N : float
            number of cycles
        ignore_limits : boolean
            ignores the upper limit of the number of cycles
            generally it should be smaller than ND_50 (=the limit of the finite life region)
            but some special evaluation methods (e.g. according to marquardt2004) require
            extrapolation to estimate an equivalent stress

        Returns
        -------
        S : float
            stress corresponding to the given number of cycles (point on the SN-curve)
        """
        return self._wc.woehler.basquin_load(N)

    def calc_N(self, S, ignore_limits=False):
        """Calculate number of cycles N for a given stress S

        Parameters
        ----------
        S : array like
            Stress (point(s) on the SN-curve)
        ignore_limits : boolean
            ignores the upper limit of the number of cycles
            generally it should be smaller than ND_50 (=the limit of the finite life region)
            but some special evaluation methods (e.g. according to marquardt2004) require
            extrapolation to estimate an equivalent stress

        Returns
        -------
        N : array like
            number of cycles corresponding to the given stress value (point on the SN-curve)
        """
        return self._wc.woehler.basquin_cycles(S)

    def calc_damage(self, loads, method="elementar", index_name="range"):
        """Calculate the damage based on the methods
         * Miner elementar (k_2 = k)
         * Miner Haibach (k_2 = 2k-1)
         * Miner original (k_2 = -inf)

        **Consider:** load collective and life curve have to be consistent:

        * range vs range
        * amplitude vs amplitude

        Parameters
        ----------
        loads : pandas series histogram
            loads (index is the load, column the cycles)
        method : str
         * 'elementar': Miner elementar (k_2 = k)
         * 'MinerHaibach': Miner Haibach (k_2 = 2k-1)
         * 'original': Miner original (k_2 = -inf)

        Returns
        -------
        damage : pd.DataFrame
            damage for every load horizont based on the load collective and the method
        """
        estimator = self._wc.fatigue

        if method == 'elementar':
            estimator = estimator.miner_elementary()
        elif method == 'MinerHaibach':
            estimator = estimator.miner_haibach()

        if index_name != "range":
            loads = loads.copy()
            names = ["range" if name == index_name else name for name in loads.index.names]
            loads.index.set_names(names, inplace=True)

        damage = estimator.damage(loads.load_collective)
        if index_name != "range":
            names = [index_name if name == "range" else name for name in loads.index.names]
            damage.index.set_names(names, inplace=True)

        return damage



# Copyright (c) 2020-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from h5py.h5t import string_dtype
from .vmap_dataset import VMAPDataset


class VMAPMetadata(VMAPDataset):
    def __init__(self, key, value):
        self._column_0 = key
        self._column_1 = value

    @property
    def attributes(self):
        return self._column_0, self._column_1

    @property
    def dtype(self):
        return string_dtype()

    @property
    def dataset_name(self):
        return 'METADATA'

    @property
    def compound_dataset(self):
        return False



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module performing rainflow counting

Overview over pyLife's rainflow counting module
-----------------------------------------------

From pyLife-2.0.0 on rainflow counting has been split into two different subtasks:

* hysteresis loop detection, done by a subclass of :class:`.AbstractDetector`.
* hysteresis loop recording, done by a subclass of :class:`.AbstractRecorder`.

That means you can combine detectors and recorders freely. You can choose
recorders and detectors that come with pyLife but also write your own custom
detectors and custom recorders.

Detectors
^^^^^^^^^

Detectors process a one dimensional time signal and detect hysteresis loops in
them. A hysteresis loop consists of the sample point where the hysteresis
starts, and the sample of the turning point where the hysteresis loop starts to
turn back towards the load level of the starting point.

Once the detector has detected such a sample pair that makes a closed
hysteresis loop it reports it to the recorder. All detectors report the load
levels, some detectors also the index to the samples defining the loop limits.

pyLife's detectors are implemented in a way that the samples are
chunkable. That means that you don't need to feed them the complete signal at
once, but you can resume the rainflow analysis later when you have the next
sample chunk.

As of now, pyLife comes with the following detectors:

* :class:`.ThreePointDetector`, classic three point algorithm, reports sample index

* :class:`.FourPointDetector`, recent four point algorithm, reports sample index

* :class:`.FKMDetector`, algorithm described by Clormann & Seeger, recommended by FKM,
  does not report sample index.


Warning
.......

Make sure you don't have any ``NaN`` values in your input signal.  They are
dropped in order to make sure not to miss any hysteresis loops and thus will
render the index invalid.  A warning is issued by :func:`~.find_turns` if
``NaN`` values are dropped.

Recorders
^^^^^^^^^

Recorders are notified by detectors about loops and will process the loop
information as they wish.

As of now, pyLife comes with the following recorders:

* :class:`.LoopValueRecorder`, only records the `from` and `to` values of all
  the closed hysteresis loops.`

* :class:`.FullRecorder`, records additionally to the `from` and `to` values
  also the indices of the loop turning points in the original time series, so
  that additional data like temperature during the loop or dwell times can be
  looked up in the original time series data.

"""

__author__ = "Johannes Mueller"
__maintainer__ = __author__

from .general import find_turns, AbstractDetector, AbstractRecorder
from .threepoint import ThreePointDetector
from .fourpoint import FourPointDetector
from .fkm import FKMDetector
from .recorders import LoopValueRecorder, FullRecorder

from .compat import RainflowCounterThreePoint, RainflowCounterFKM

__all__ = [
    'find_turns',
    'AbstractDetector',
    'AbstractRecorder',
    'ThreePointDetector',
    'FourPointDetector',
    'FKMDetector',
    'LoopValueRecorder',
    'FullRecorder',
    'RainflowCounterThreePoint',
    'RainflowCounterFKM',
]



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys, os, copy
import warnings
import pytest

import numpy as np
import pandas as pd
import numpy.testing as testing

import pylife.strength.meanstress as MST


def test_haigh_diagram_fail_no_meanstress_sens_index():
    hd = pd.Series(dtype=np.float64)
    with pytest.raises(AttributeError, match="A Haigh Diagram needs an index level 'R'."):
        hd.haigh_diagram


def test_haigh_diagram_fail_no_interval_index():
    hd = pd.Series([], index=pd.Index([], name='R'), dtype=np.float64)
    with pytest.raises(AttributeError, match="The 'R' index must be an IntervalIndex."):
        hd.haigh_diagram


def test_haigh_diagram_multi_index_fail_no_interval_index():
    hd = pd.Series([], index=pd.MultiIndex.from_frame(pd.DataFrame([], columns=['foo', 'R'])), dtype=np.float64)
    with pytest.raises(AttributeError, match="The 'R' index must be an IntervalIndex."):
        hd.haigh_diagram


def test_haigh_diagram_from_dict_empty():
    hd = MST.HaighDiagram.from_dict({})
    expected = pd.Series([], index=pd.IntervalIndex([], name='R'), dtype=np.float64)
    pd.testing.assert_series_equal(hd.to_pandas(), expected)


def test_haigh_diagram_from_dict_filled():
    hd = MST.HaighDiagram.from_dict({
        (1.0, np.inf): 0.0,
        (-np.inf, 0.0): 0.5,
        (0.0, 1.0): 0.167
    })

    expected_index = pd.IntervalIndex.from_tuples([(1.0, np.inf), (-np.inf, 0.0), (0.0, 1.0)], name='R')
    expected = pd.Series([0.0, 0.5, 0.167], index=expected_index)
    pd.testing.assert_series_equal(hd.to_pandas(), expected)


def test_haigh_diagram_fkm_goodman_single_M():
    hd = MST.HaighDiagram.fkm_goodman(pd.Series({'M': 0.5}))

    expected_index = pd.IntervalIndex.from_tuples([(1.0, np.inf), (-np.inf, 0.0), (0.0, 1.0)], name='R')
    expected = pd.Series([0.0, 0.5, 0.166667], index=expected_index)
    pd.testing.assert_series_equal(hd.to_pandas(), expected, rtol=1e-4)


def test_haigh_diagram_fkm_goodman_single_M_M2():
    hd = MST.HaighDiagram.fkm_goodman(pd.Series({'M': 0.4, 'M2': 0.15}))

    expected_index = pd.IntervalIndex.from_tuples([(1.0, np.inf), (-np.inf, 0.0), (0.0, 1.0)], name='R')
    expected = pd.Series([0.0, 0.4, 0.15], index=expected_index)
    pd.testing.assert_series_equal(hd.to_pandas(), expected, rtol=1e-4)


def test_haigh_diagram_fkm_goodman_multiple_M():
    hd = MST.HaighDiagram.fkm_goodman(pd.DataFrame({
        'M': [0.5, 0.4]
    }, index=pd.Index([1, 2], name='element_id')))

    expected_index = pd.MultiIndex.from_tuples([
        (1, pd.Interval(1.0, np.inf)),
        (1, pd.Interval(-np.inf, 0.0)),
        (1, pd.Interval(0.0, 1.0)),
        (2, pd.Interval(1.0, np.inf)),
        (2, pd.Interval(-np.inf, 0.0)),
        (2, pd.Interval(0.0, 1.0)),
    ], names=['element_id', 'R'])
    expected = pd.Series([0.0, 0.5, 0.166667, 0.0, 0.4, 0.13333], index=expected_index)
    pd.testing.assert_series_equal(hd.to_pandas(), expected, rtol=1e-4)


def test_haigh_diagram_fkm_goodman_sized_M():
    MST.HaighDiagram.fkm_goodman(pd.DataFrame({
        'M': 0.5
    }, index=pd.Index(np.arange(101), name='element_id')))


def test_haigh_diagram_fkm_goodman_multiple_M_M2():
    hd = MST.HaighDiagram.fkm_goodman(pd.DataFrame({
        'M': [0.5, 0.4],
        'M2': [0.2, 0.15]
    }, index=pd.Index([1, 2], name='element_id')))

    expected_index = pd.MultiIndex.from_tuples([
        (1, pd.Interval(1.0, np.inf)),
        (1, pd.Interval(-np.inf, 0.0)),
        (1, pd.Interval(0.0, 1.0)),
        (2, pd.Interval(1.0, np.inf)),
        (2, pd.Interval(-np.inf, 0.0)),
        (2, pd.Interval(0.0, 1.0)),
    ], names=['element_id', 'R'])
    expected = pd.Series([0.0, 0.5, 0.2, 0.0, 0.4, 0.15], index=expected_index)
    pd.testing.assert_series_equal(hd.to_pandas(), expected, rtol=1e-4)


## TODO
##
## * left closed intervals?

# @pytest.mark.parametrize('left, right', [(-np.inf, 0.1), (0.0, 1.0), (-np.inf, np.inf)])
# def test_haigh_diagram_fail_R_range_not_covered(left, right):
#     hd = pd.Series([0.5], index=pd.IntervalIndex.from_tuples([(left, right)], name='R'), dtype=np.float64)
#     with pytest.raises(AttributeError, match=r"The 'R' IntervalIndex must cover the exact range \(-inf, 1.0\]."):
#         hd.haigh_diagram


# def test_haigh_diagram_multi_index_interval_index_fail_R_range_not_covered():
#     foo_level = pd.Index([], name='foo')
#     R_level = pd.IntervalIndex([], name='R')
#     hd = pd.Series([], index=pd.MultiIndex.from_product([foo_level, R_level]), dtype=np.float64)
#     with pytest.raises(AttributeError, match=r"The 'R' IntervalIndex must cover the exact range \(-inf, 1.0\]."):
#         hd.haigh_diagram


def test_haigh_diagram_fail_R_is_overlapping_multi():
    idx = pd.MultiIndex.from_tuples([
        (1, pd.Interval(1.0, np.inf)),
        (1, pd.Interval(-np.inf, -1.0)),
        (1, pd.Interval(-1.0, 0.0)),
        (2, pd.Interval(1.0, np.inf)),
        (2, pd.Interval(-np.inf, -0.5)),
        (2, pd.Interval(-0.6, 0.0)),
    ], names=['element_id', 'R'])
    hd = pd.Series(1.0, index=idx)
    with pytest.raises(AttributeError, match=r"The intervals of the 'R' IntervalIndex must not overlap."):
        hd.haigh_diagram


def test_haigh_diagram_fail_R_is_overlapping():
    idx = pd.IntervalIndex.from_tuples([(-np.inf, 0.1), (0.0, 1.0)], name='R')
    hd = pd.Series([0.5, 0.5], index=idx)
    with pytest.raises(AttributeError, match=r"The intervals of the 'R' IntervalIndex must not overlap."):
        hd.haigh_diagram


def test_haigh_diagram_fail_if_R_has_gaps():
    idx = pd.IntervalIndex.from_tuples([(-np.inf, 0.0), (0.1, 1.0)], name='R')
    hd = pd.Series([0.5, 0.5], index=idx)
    with pytest.raises(AttributeError, match=r"The intervals of the 'R' IntervalIndex must not have gaps."):
        hd.haigh_diagram


def test_haigh_diagram_fail_if_R_has_gaps_multi():
    idx = pd.MultiIndex.from_tuples([
        (1, pd.Interval(1.0, np.inf)),
        (1, pd.Interval(-np.inf, -1.0)),
        (1, pd.Interval(-1.0, 0.0)),
        (2, pd.Interval(1.0, np.inf)),
        (2, pd.Interval(-np.inf, -0.5)),
        (2, pd.Interval(-0.4, 0.0)),
    ], names=['element_id', 'R'])
    hd = pd.Series(1.0, index=idx)
    with pytest.raises(AttributeError, match=r"The intervals of the 'R' IntervalIndex must not have gaps."):
        hd.haigh_diagram


def test_haigh_diagram_multiindex_unique():
    foo_level = pd.Index([1, 2, 3], name='foo')
    R_level = pd.IntervalIndex.from_tuples([(1., np.inf), (-np.inf, 1.)], name='R')
    hd = pd.Series([1.]*6, index=pd.MultiIndex.from_product([foo_level, R_level]), dtype=np.float64)
    pd.testing.assert_index_equal(hd.haigh_diagram._R_index, R_level)


def test_haigh_diagram():
    hd = pd.Series([0.5], index=pd.IntervalIndex.from_tuples([(-np.inf, 1.0)], name='R'))
    hd.haigh_diagram


@pytest.mark.parametrize('M, from_val, to_val, R_goal, expected_range, expected_mean', [
    (0.5, -1., 1., -1.0, 2.0, 0.0),
    (0.5, -0.4, 1.2, -1.0, 2.0, 0.0),
    (0.5, 0., 4./3., -1.0, 2.0, 0.0),
    (0.5, -1., 1., 0.0, 4./3., 2./3.),
    (0.5, 1., -1., -1./3., 1.6, 0.4),
    (0.25, -0.4, 1.2, -1.0, 1.8, 0.0),
    (0.25, -0.9, 0.9, -1./3., 1.6, 0.4),
    (0.5/3., 7./12., 21./12., 0.0, 4./3., 2./3.)
])
def test_haigh_diagram_one_segment(M, from_val, to_val, R_goal, expected_range, expected_mean):
    hd = pd.Series([M], index=pd.IntervalIndex.from_tuples([(-np.inf, 1.0)], name='R'))
    cycle = pd.DataFrame({
        'from': [from_val],
        'to': [to_val]
    })

    res = hd.haigh_diagram.transform(cycle, R_goal)
    expected = pd.DataFrame({'range': [expected_range], 'mean': [expected_mean]})
    pd.testing.assert_frame_equal(res, expected)


@pytest.mark.parametrize('M, from_val, to_val, R_goal, expected_range, expected_mean', [
    (0.5, -1., 1., -1.0, 2.0, 0.0),
    (0.5, -0.4, 1.2, -1.0, 2.0, 0.0),
    (0.5, 0., 4./3., -1.0, 2.0, 0.0),
    (0.5, -1., 1., 0.0, 4./3., 2./3.),
    (0.5, 1., -1., -1./3., 1.6, 0.4),
    (0.25, -0.4, 1.2, -1.0, 1.8, 0.0),
    (0.25, -0.9, 0.9, -1./3., 1.6, 0.4),
    (0.5, 7./12., 21./12., -1.0, 2.0, 0.0),
    (0.5, -1., 1., 1./3., 7./6., 7./6.)
])
def test_haigh_diagram_two_segment(M, from_val, to_val, R_goal, expected_range, expected_mean):
    idx = pd.IntervalIndex.from_tuples([
        (-np.inf, 0.0),
        (0.0, 1.0)
    ], name='R')
    hd = pd.Series([M, M/3.], index=idx)

    cycle = pd.DataFrame({
        'from': [from_val],
        'to': [to_val]
    })

    res = hd.haigh_diagram.transform(cycle, R_goal)
    expected = pd.DataFrame({'range': [expected_range], 'mean': [expected_mean]})
    pd.testing.assert_frame_equal(res, expected)


def test_haigh_diagram_transform_to_inf():
    idx = pd.IntervalIndex.from_tuples([
        (-np.inf, 0.0),
        (0.0, 1.0)
    ], name='R')
    hd = pd.Series([0., 0.], index=idx)

    cycle = pd.DataFrame({
        'from': [-1.],
        'to': [1.]
    })

    res = hd.haigh_diagram.transform(cycle, R_goal=-np.inf)
    expected = pd.DataFrame({'range': [2.], 'mean': [-1.]})
    pd.testing.assert_frame_equal(res, expected)


def test_haigh_diagram_transform_to_R_gt_1():
    idx = pd.IntervalIndex.from_tuples([
        (1.0, np.inf),
        (-np.inf, 0.0),
    ], name='R')

    hd = pd.Series([0.5, 0.5], index=idx)

    cycle = pd.DataFrame({
        'from': [-1.],
        'to': [1.]
    })

    res = hd.haigh_diagram.transform(cycle, R_goal=5.0)

    expected = pd.DataFrame({'range': [8.], 'mean': [-6.]})
    pd.testing.assert_frame_equal(res, expected)


def test_haigh_diagram_broadcast_constant_intervals():
    R_intervals = pd.IntervalIndex.from_tuples([
        (1.0, np.inf),
        (-np.inf, 0.0),
    ], name='R')

    elements = pd.Index([1, 2], name='element_id')

    idx = pd.MultiIndex.from_product([elements, R_intervals])

    hd = pd.Series([0.0, 0.5, 0.0, 1./3.], index=idx)

    cycles = pd.DataFrame({
        'from': [-1., -2.],
        'to': [1., 2.]
    }, index=pd.Index([10, 11], name='cycle_number'))

    expected = pd.DataFrame({
        'range': [4.0, 8.0, 3.0, 6.0],
        'mean': [-2.0, -4.0, -1.5, -3.0]
    }, index=pd.MultiIndex.from_tuples([
        (1, 10), (1, 11),
        (2, 10), (2, 11)
    ], names=['element_id', 'cycle_number']))

    res = hd.haigh_diagram.transform(cycles, R_goal=-np.inf)

    pd.testing.assert_frame_equal(res, expected)


def test_haigh_diagram_broadcast_variable_intervals():
    R_intervals = pd.IntervalIndex.from_tuples([
        (1.0, np.inf),
        (-np.inf, 0.0),
    ], name='R')

    elements = pd.Index([1, 2], name='element_id')

    idx = pd.MultiIndex.from_product([elements, R_intervals])

    idx = pd.MultiIndex.from_tuples([
        (1, pd.Interval(1.0, np.inf)),
        (1, pd.Interval(-np.inf, -1.0)),
        (1, pd.Interval(-1.0, 0.0)),
        (2, pd.Interval(1.0, np.inf)),
        (2, pd.Interval(-np.inf, -0.5)),
        (2, pd.Interval(-0.5, 0.0)),
    ], names=['element_id', 'R'])

    hd = pd.Series([0.0, 0.5, 0.5, 0.0, 1./3., 1./3.], index=idx)

    cycles = pd.DataFrame({
        'from': [-1., -2.],
        'to': [1., 2.]
    }, index=pd.Index([10, 11], name='cycle_number'))

    expected = pd.DataFrame({
        'range': [4.0, 8.0, 3.0, 6.0],
        'mean': [-2.0, -4.0, -1.5, -3.0]
    }, index=pd.MultiIndex.from_tuples([
        (1, 10), (1, 11),
        (2, 10), (2, 11)
    ], names=['element_id', 'cycle_number']))

    res = hd.haigh_diagram.transform(cycles, R_goal=-np.inf)
    pd.testing.assert_frame_equal(res, expected)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
from scipy.stats import norm
import scipy.integrate as integrate


class FailureProbability:
    '''Strength representation to calculate failure probabilities

    The strength is represented as a log normal distribution of
    strength_median and strength_std.

    Failure probabilities can be calculated for a given load or load
    distribution.

    Parameters
    ----------
    strength_median : array_like, shape (N, )
        The median value of the strength
    strength_std : array_like, shape (N, )
        The standard deviation of the strength

    Note
    ----
    We assume that the load and the strength are statistically
    distributed values. In case the load is higher than the strength
    we get failure. So if we consider a quantile of our load
    distribution of a probability p_load, the probability of failure
    due to a load of this quantile is p_load times the probability
    that the strength lies within this quantile or below.

    So in order to calculate the total failure probability, we need to
    integrate the load's pdf times the strength' cdf from -inf to +inf.

    '''

    def __init__(self, strength_median, strength_std):
        self.s_50 = np.log10(strength_median)
        self.s_std = strength_std

    def pf_simple_load(self, load):
        '''Failure probability for a simple load value

        Parameters
        ----------
        load : array_like, shape (N,) consistent with class parameters
            The load of for which the failure probability is
            calculated.

        Returns
        -------
        failure probability : numpy.ndarray or float

        Notes
        -----
        This is the case of a non statistical load. So failure occurs
        if the strength is below the given load, i.e. the strength'
        cdf at the load.
        '''
        return norm.cdf(np.log10(load), loc=self.s_50, scale=self.s_std)

    def pf_norm_load(self, load_median, load_std, lower_limit=None, upper_limit=None):
        '''Failure probability for a log normal distributed load

        Parameters
        ----------
        load_median : array_like, shape (N,) consistent with class parameters
            The median of the load distribution for which the failure
            probability is calculated.
        load_std : array_like, shape (N,) consistent with class parameters
            The standard deviation of the load distribution
        lower_limit : float, optional
            The lower limit of the integration, default None
        upper_limit : float, optional
            The upper limit of the integration, default None

        Returns
        -------
        failure probability : numpy.ndarray or float

        Notes
        -----

        The log normal distribution of the load is determined by the
        load parameters. Only load distribution between
        ``lower_limit`` and ``upper_limit`` is considered.

        For small values for ``load_std`` this function gives the same
        result as ``pf_simple_load``.

        Note
        ----
        The load and strength distributions are transformed in a way,
        that the median of the load distribution is zero. This
        guarantees that in any case we can provide a set of relevant
        points to take into account for the integration.

        '''
        lm = np.log10(load_median)

        sc = load_std

        if lower_limit is None:
            lower_limit = -16.*sc
        else:
            lower_limit -= lm
        if upper_limit is None:
            upper_limit = +16.*sc
        else:
            upper_limit -= lm

        q1, err_est = integrate.quad(
            lambda x: norm.pdf(x, loc=0.0, scale=sc) * norm.cdf(x, loc=self.s_50-lm, scale=self.s_std),
            lower_limit, upper_limit)

        return q1

    def pf_arbitrary_load(self, load_values, load_pdf):
        ''' Calculates the failure probability for an arbitrary load

        Parameters
        ----------
        load_values : array_like, shape (N,)
            The load values of the load distribution
        load_pdf : array_like, shape (N, )
            The probability density values for the ``load_value`` values to
            occur

        Returns
        -------
        failure probability : numpy.ndarray or float
        '''
        if load_values.shape != load_pdf.shape:
            raise ValueError("Load values and pdf must have same dimensions.")

        strength_cdf = norm.cdf(load_values, loc=self.s_50, scale=self.s_std)


        # TODO: remove this once we drop python 3.8 support
        trapezoid = np.trapezoid if hasattr(np, 'trapezoid') else np.trapz

        return trapezoid(load_pdf * strength_cdf, x = load_values)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import numpy as np
import pandas as pd

from pylife.core import *

foo_bar_baz = pd.DataFrame({'foo': [1.0, 1.0], 'bar': [1.0, 1.0], 'baz': [1.0, 1.0]})

def test_keys_dataframe():
    pd.testing.assert_index_equal(foo_bar_baz.test_accessor_none.keys(), pd.Index(['foo', 'bar', 'baz']))


def test_keys_series():
    pd.testing.assert_index_equal(foo_bar_baz.iloc[0].test_accessor_none.keys(), pd.Index(['foo', 'bar', 'baz']))

def test_missing_keys_none():
    assert foo_bar_baz.test_accessor_none.get_missing_keys(['foo', 'bar']) == []


def test_missing_keys_one():
    assert foo_bar_baz.test_accessor_none.get_missing_keys(['foo', 'foobar']) == ['foobar']


def test_missing_keys_two():
    assert set(foo_bar_baz.test_accessor_none.get_missing_keys(['foo', 'foobar', 'barfoo'])) == set(['foobar', 'barfoo'])


def test_from_parameters_frame():
    foo = [1.0, 2.0, 3.0]
    bar = [10.0, 20.0, 30.0]
    baz = [11.0, 12.0, 13.0]
    accessor = AccessorNone.from_parameters(foo=foo, bar=bar, baz=baz)
    pd.testing.assert_index_equal(accessor.keys(), pd.Index(['foo', 'bar', 'baz']))
    expected_obj = pd.DataFrame({'foo': foo, 'bar': bar, 'baz': baz})
    pd.testing.assert_frame_equal(accessor._obj, expected_obj)
    assert accessor.some_property == 42


def test_from_parameters_series_columns():
    foo = 1.0
    bar = 10.0
    baz = 11.0
    accessor = AccessorNone.from_parameters(foo=foo, bar=bar, baz=baz)
    pd.testing.assert_index_equal(accessor.keys(), pd.Index(['foo', 'bar', 'baz']))
    expected_obj = pd.Series({'foo': foo, 'bar': bar, 'baz': baz})
    pd.testing.assert_series_equal(accessor._obj, expected_obj)
    assert accessor.some_property == 42


def test_from_parameters_series_index():
    foo = [1.0, 2.0, 3.0]
    accessor = AccessorOneDim.from_parameters(foo=foo)
    pd.testing.assert_index_equal(accessor.keys(), pd.Index(['foo']))
    expected_obj = pd.Series({'foo': foo})
    pd.testing.assert_series_equal(accessor._obj, expected_obj)
    assert accessor.some_property == 42


def test_from_parameters_missing_keys():
    foo = 1.0
    baz = 10.0
    with pytest.raises(AttributeError, match=r'^AccessorNone.*bar'):
        AccessorNone.from_parameters(foo=foo, baz=baz)


@pd.api.extensions.register_series_accessor('test_accessor_one_dim')
class AccessorOneDim(PylifeSignal):
    def _validate(self):
        if not isinstance(self._obj, pd.Series):
            raise TypeError("This accessor takes only pd.Series")

    @property
    def some_property(self):
        return 42

@pd.api.extensions.register_series_accessor('test_accessor_none')
@pd.api.extensions.register_dataframe_accessor('test_accessor_none')
class AccessorNone(PylifeSignal):
    def _validate(self):
        self.fail_if_key_missing(['foo', 'bar'])

    def already_here(self):
        return 23

    @property
    def some_property(self):
        return 42

    @property
    def not_working_property(self):
        self._missing_attribute


@pd.api.extensions.register_series_accessor('test_accessor_one')
@pd.api.extensions.register_dataframe_accessor('test_accessor_one')
class AccessorOne(PylifeSignal):
    def _validate(self):
        self.fail_if_key_missing(['foo', 'foobar'])


@pd.api.extensions.register_dataframe_accessor('test_accessor_two')
class AccessorTwo(PylifeSignal):
    def _validate(self):
        self.fail_if_key_missing(['foo', 'foobar', 'barfoo'])


def test_signal_broadcast_inheritance_series():
    assert isinstance(foo_bar_baz.loc[0].test_accessor_none, Broadcaster)


def test_signal_broadcast_inheritance_frame():
    assert isinstance(foo_bar_baz.test_accessor_none, Broadcaster)


def test_fail_if_missing_keys_none():
    foo_bar_baz.test_accessor_none


def test_fail_if_missing_keys_one_dataframe():
    with pytest.raises(AttributeError, match=r'^AccessorOne.*foobar'):
        foo_bar_baz.test_accessor_one


def test_fail_if_missing_keys_one_series():
    with pytest.raises(AttributeError, match=r'^AccessorOne.*foobar'):
        foo_bar_baz.loc[0].test_accessor_one


def test_fail_if_missing_keys_two():
    with pytest.raises(AttributeError, match=r'^AccessorTwo.*(foobar|barfoo).*(barfoo|foobar)'):
        foo_bar_baz.test_accessor_two


def test_register_method():
    @register_method(AccessorNone, 'foo_method')
    def foo(df):
        return pd.DataFrame({'baz': df['foo'] + df['bar']})

    accessor = foo_bar_baz.test_accessor_none
    pd.testing.assert_frame_equal(accessor.foo_method(), pd.DataFrame({'baz': [2.0, 2.0]}))


def test_getattr_no_method():
    accessor = foo_bar_baz.test_accessor_none
    assert accessor.already_here() == 23
    assert accessor.some_property == 42


def test_register_method_missing_attribute():
    @register_method(AccessorNone, 'another_method')
    def foo(df):
        return pd.DataFrame({'baz': df['foo'] + df['bar']})

    accessor = foo_bar_baz.test_accessor_none
    with pytest.raises(AttributeError, match=r'^\'AccessorNone\' object has no attribute \'_missing_attribute\''):
        accessor.not_working_property


def test_register_method_fail_duplicate():
    with pytest.raises(ValueError, match=r'^Method \'bar_method\' already registered in AccessorNone'):
        @register_method(AccessorNone, 'bar_method')
        def bar1(df):
            return pd.DataFrame({'baz': df['foo'] + df['bar']})

        @register_method(AccessorNone, 'bar_method')
        def bar2(df):
            return pd.DataFrame({'baz': df['foo'] - df['bar']})


def test_register_method_fail_already():
    with pytest.raises(ValueError, match=r'^AccessorNone already has an attribute \'already_here\''):
        @register_method(AccessorNone, 'already_here')
        def already_here_method(df):
            return pd.DataFrame({'baz': df['foo'] + df['bar']})


def test_pandas_series():
    series = foo_bar_baz.iloc[0]
    pd.testing.assert_series_equal(series.test_accessor_none.to_pandas(), series)


def test_pandas_frame():
    pd.testing.assert_frame_equal(foo_bar_baz.test_accessor_none.to_pandas(), foo_bar_baz)



import pytest

import time
import numpy as np


@pytest.fixture
def benchmarker():

    def the_benchmarker(rainflow_counter):
        load_signal = np.loadtxt('load_signal.csv')

        tic = time.perf_counter()

        rainflow_counter.process(load_signal)

        toc = time.perf_counter()
        elapsed = toc - tic

        classname = rainflow_counter.__class__.__name__
        print(f"Processing {classname} took {elapsed:0.4f} seconds")

        return elapsed

    return the_benchmarker



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

import numpy as np
from scipy import stats


from pylife.utils.functions import scattering_range_to_std, std_to_scattering_range


class Likelihood:
    """Calculate the likelihood a fatigue dataset matches with Wöhler curve parameters."""

    def __init__(self, fatigue_data):
        self._fd = fatigue_data

    def likelihood_total(self, SD, TS, k_1, ND, TN):
        """
        Produces the likelihood functions that are needed to compute the parameters of the woehler curve.
        The likelihood functions are represented by probability and cummalative distribution functions.
        The likelihood function of a runout is 1-Li(fracture). The functions are added together, and the
        negative value is returned to the optimizer.

        Parameters
        ----------
        SD:
            Endurnace limit start value to be optimzed, unless the user fixed it.
        TS:
            The scatter in load direction 1/TS to be optimzed, unless the user fixed it.
        k_1:
            The slope k_1 to be optimzed, unless the user fixed it.
        ND:
            Load-cycle endurance start value to be optimzed, unless the user fixed it.
        TN:
            The scatter in load-cycle direction 1/TN to be optimzed, unless the user fixed it.

        Returns
        -------
        neg_sum_lolli :
            Sum of the log likelihoods. The negative value is taken since optimizers in statistical
            packages usually work by minimizing the result of a function. Performing the maximum likelihood
            estimate of a function is the same as minimizing the negative log likelihood of the function.

        """
        return self.likelihood_finite(SD, k_1, ND, TN) + self.likelihood_infinite(SD, TS)

    def likelihood_finite(self, SD, k_1, ND, TN):
        if not (SD > 0.0).all():
            return -np.inf
        fractures = self._fd.fractures
        x = np.log10(fractures.cycles * ((fractures.load/SD)**k_1))
        mu = np.log10(ND)
        std_log = scattering_range_to_std(TN)
        log_likelihood = np.log(stats.norm.pdf(x, mu, std_log))

        return log_likelihood.sum()

    def likelihood_infinite(self, SD, TS):
        """
        Produces the likelihood functions that are needed to compute the endurance limit and the scatter
        in load direction. The likelihood functions are represented by a cummalative distribution function.
        The likelihood function of a runout is 1-Li(fracture).

        Parameters
        ----------
        SD:
            Endurnace limit start value to be optimzed, unless the user fixed it.
        TS:
            The scatter in load direction 1/TS to be optimzed, unless the user fixed it.

        Returns
        -------
        neg_sum_lolli :
            Sum of the log likelihoods. The negative value is taken since optimizers in statistical
            packages usually work by minimizing the result of a function. Performing the maximum likelihood
            estimate of a function is the same as minimizing the negative log likelihood of the function.

        """
        infinite_zone = self._fd.infinite_zone
        std_log = scattering_range_to_std(TS)
        t = np.logical_not(self._fd.infinite_zone.fracture).astype(np.float64)
        likelihood = stats.norm.cdf(np.log10(infinite_zone.load/SD),  scale=abs(std_log))
        non_log_likelihood = t+(1.-2.*t)*likelihood
        if non_log_likelihood.eq(0.0).any():
            return -np.inf

        return np.log(non_log_likelihood).sum()



# Copyright (c) 2019-2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
from pylife.rainflow_ext import threepoint_loop

from .general import AbstractDetector


class ThreePointDetector(AbstractDetector):
    r"""Classic three point rainflow counting algorithm.

    .. jupyter-execute::

        from pylife.stress.timesignal import TimeSignalGenerator
        import pylife.stress.rainflow as RF

        ts = TimeSignalGenerator(10, {
            'number': 50,
            'amplitude_median': 1.0, 'amplitude_std_dev': 0.5,
            'frequency_median': 4, 'frequency_std_dev': 3,
            'offset_median': 0, 'offset_std_dev': 0.4}, None, None).query(10000)

        rfc = RF.ThreePointDetector(recorder=RF.LoopValueRecorder())
        rfc.process(ts)

        rfc.recorder.collective

    Alternatively you can ask the recorder for a histogram matrix:

    .. jupyter-execute::

        rfc.recorder.histogram(bins=16)

    We take three turning points into account to detect closed hysteresis loops.

    * start: the point where the loop is starting from
    * front: the turning point after the start
    * back: the turning point after the front

    A loop is considered closed if following conditions are met:

    * the load difference between front and back is bigger than or
      equal the one between start and front. In other words: if the
      back goes beyond the starting point. For example (A-B-C) and
      (B-C-D) not closed, whereas (C-D-E) is.

    * the loop init has not been a loop front in a prior closed
      loop. For example F would close the loops (D-E-F) but D is
      already front of the closed loop (C-D-E).

    * the load level of the front has already been covered by a prior
      turning point. Otherwise it is considered part of the front
      residuum.

    When a loop is closed it is possible that the loop back also
    closes unclosed loops of the past by acting as loop back for an
    unclosed start/front pair. For example E closes the loop (C-D-E)
    and then also (A-B-E).

    ::

        Load -----------------------------
        |        x B               F x
        --------/-\-----------------/-----
        |      /   \   x D         /
        ------/-----\-/-\---------/-------
        |    /     C x   \       /
        --\-/-------------\-----/---------
        |  x A             \   /
        --------------------\-/-----------
        |                    x E
        ----------------------------------
        |              Time

    .. _subsection_TP: ../demos/rainflow.ipynb#Classic-Three-Point-Counting
    """

    def __init__(self, recorder):
        """Instantiate a ThreePointDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        super().__init__(recorder)

    def process(self, samples):
        """Process a sample chunk.

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : ThreePointDetector
            The ``self`` object so that processing can be chained
        """
        samples = np.asarray(samples)

        if len(self._residuals) == 0:
            residuals = samples[:1]
        else:
            residuals = self._residuals[:-1]

        turns_index, turns_values = self._new_turns(samples)

        turns = np.concatenate((residuals, turns_values, samples[-1:]))
        turns_index = np.concatenate((self._residual_index, turns_index.astype(np.uintp)))

        highest_front = np.argmax(residuals)
        lowest_front = np.argmin(residuals)

        (
            from_vals,
            to_vals,
            from_index,
            to_index,
            residual_index
        ) = threepoint_loop(turns, turns_index, highest_front, lowest_front, len(residuals))

        self._recorder.record_values(from_vals, to_vals)
        self._recorder.record_index(from_index, to_index)

        self._residuals = turns[residual_index]
        self._residual_index = turns_index[residual_index[:-1]]
        self._recorder.report_chunk(len(samples))

        return self



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module for time signal handling

Warning
-------

This module is not considered finalized even though it is part of pylife-2.0.
Breaking changes might occur in upcoming minor releases.
"""

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.signal as signal

try:
    import tsfresh as ts

    _HAVE_TSFRESH = True
except ModuleNotFoundError:
    _HAVE_TSFRESH = False


class TimeSignalGenerator:
    r"""Generates mixed time signals

    The generated time signal is a mixture of random sets of sinus signals

    For each set the user supplys a dict describing the set::

      sinus_set = {
          'number': number of signals
          'amplitude_median':
          'amplitude_std_dev':
          'frequency_median':
          'frequency_std_dev':
          'offset_median':
          'offset_std_dev':
      }

    The amplitudes (:math:`A`), fequencies (:math:`\omega`) and
    offsets (:math:`c`) are then norm distributed. Each sinus signal
    looks like

            :math:`s = A \sin(\omega t + \phi) + c`

    where :math:`phi` is a random value between 0 and :math:`2\pi`.

    So the whole sinus :math:`S` set is given by the following expression:

            :math:`S = \sum^n_i A_i \sin(\omega_i t + \phi_i) + c_i`.
    """

    def __init__(self, sample_rate, sine_set, gauss_set, log_gauss_set):
        sine_amplitudes = stats.norm.rvs(
            loc=sine_set["amplitude_median"],
            scale=sine_set["amplitude_std_dev"],
            size=sine_set["number"],
        )
        sine_frequencies = stats.norm.rvs(
            loc=sine_set["frequency_median"],
            scale=sine_set["frequency_std_dev"],
            size=sine_set["number"],
        )
        sine_offsets = stats.norm.rvs(
            loc=sine_set["offset_median"],
            scale=sine_set["offset_std_dev"],
            size=sine_set["number"],
        )
        sine_phases = 2.0 * np.pi * np.random.rand(sine_set["number"])

        self.sine_set = list(
            zip(sine_amplitudes, sine_frequencies, sine_phases, sine_offsets)
        )

        self.sample_rate = sample_rate
        self.time_position = 0.0

    def query(self, sample_num):
        """Gets a sample chunk of the time signal

        Parameters
        ----------
        sample_num : int
            number of the samples requested

        Returns
        -------
        samples : 1D numpy.ndarray
            the requested samples


        You can query multiple times, the newly delivered samples
        will smoothly attach to the previously queried ones.
        """
        samples = np.zeros(sample_num)
        end_time_position = self.time_position + (sample_num - 1) / self.sample_rate

        for ampl, omega, phi, offset in self.sine_set:
            periods = np.floor(self.time_position / omega)
            start = self.time_position - periods * omega
            end = end_time_position - periods * omega
            time = np.linspace(start, end, sample_num)
            samples += ampl * np.sin(omega * time + phi) + offset

        self.time_position = end_time_position + 1.0 / self.sample_rate

        return samples

    def reset(self):
        """Resets the generator

        A resetted generator behaves like a new generator.
        """
        self.time_position = 0.0


def fs_calc(df):
    """
    Calculates the sample frequency of a DataFrame time series

    Parameters
    ----------
    df : DataFrame
        time series.

    Returns
    -------
    fs : int, float
        sample freqency

    """
    try:
        fs = 1 / np.mean(np.diff(df.index))
    except TypeError:
        print("Index has to be a number not a string. We assume fs = 1")
        fs = 1
    return fs


def resample_acc(df, fs=1):
    """Resamples a pandas time series DataFrame

    Parameters
    ----------
    df: DataFrame

    time_col: str
        column name of the time column
    fs: float
        sample rate of the resampled time series

    Returns
    -------
    DataFrame
    """
    index_new = np.arange(df.index.min(), df.index.max() + 1 / fs, 1 / fs)

    df_rs = pd.DataFrame(
        df.apply(lambda x: np.interp(index_new, df.index, x)).values,
        index=index_new,
        columns=df.columns,
    )
    return df_rs


def butter_bandpass(df, lowcut, highcut, order=5):
    """Use the functonality of scipy


    Parameters
    ----------

    df: DataFrame
    lowcut : float
        low frequency
    highcut : float
        high freqency.
    order : int, optional
        Butterworth filter order. The default is 5.

    Returns
    -------
    TSout : DataFrame

    """
    fs = fs_calc(df)
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = signal.butter(order, [low, high], btype="bandpass")
    return df.apply(lambda x: signal.filtfilt(b, a, x, padlen=int(fs / 2)))


def psd_df(df_ts, nfft=512, nperseg=256):
    """
    calculates the psd using Welch algorithm from matplotlib functionality

    Parameters
    ----------
    df_ts : DataFram
        Time series dataframe
    nfft : int, optional
        Length of the FFT. The default is 512.

    Returns
    -------
    df_psd : DataFrame
        PSD.

    """

    nperseg = min(nperseg, nfft)
    fs = fs_calc(df_ts)
    df_psd = pd.DataFrame()
    for col in df_ts:
        freq, df_psd[col] = signal.welch(df_ts[col], fs=fs, nperseg=nperseg, nfft=nfft)
    df_psd.index = pd.Index(freq, name="frequency")
    return df_psd


def _prepare_rolling(df):
    """
    Adds ID, time to the dataset for TsFresh, We would need different ID's if we had
    independant timeseries -like timeseries for different robots.

    Parameters
    ----------
    df: pandas DataFrame
        input data
    self : TimeSignalPrep class


    Returns
    -------
    df : pandas DataFrame
        output DataFrame with added id, time

    """
    prep_roll = df.copy()
    prep_roll["id"] = 0
    prep_roll["time"] = df.index.values
    prep_roll["time"] = prep_roll["time"].subtract(prep_roll["time"].values[0])
    prep_roll.index = prep_roll["time"]

    return prep_roll


def _roll_dataset(prep_roll_df, window_size=1000, overlap=200):
    """
    Rolls dataset in windows so we can later extract features from every window
    Parameters
    ----------
    prep_roll: output from prepare_rolling

    window_size : int , optional
         window size of the rolled segments  -the default is 1000.
    overlap : int, optional
         overlap between 2 adjecent windows -The default is 200.

    Returns
    -------
    df_rolled : pandas DataFrame
        rolled DataFrame

    """

    # Create Rolled Dataset with Parameter rolling_direction & window_size
    # throws away the last halfshift
    rolling_direction = window_size - overlap
    cycles = int((len(prep_roll_df) - window_size) / rolling_direction) + 1

    parts = []
    # shiften
    for i in range(cycles):
        position = (rolling_direction) * i
        shift = prep_roll_df.iloc[position : position + window_size, :].copy()
        # change IDs to format (id,time)
        df = pd.DataFrame(
            {
                "id": np.int64(np.zeros(len(shift), dtype=int)),
                "max_time": shift.iloc[-1, -1],
            }
        )

        shift["id"] = pd.MultiIndex.from_frame(df).to_numpy()

        parts.append(shift)

    return pd.concat(parts, ignore_index=True)


def _extract_feature_df(df_rolled, feature="maximum"):
    """Extracts features like "abs_energy" or "maximum" from the rolled dataset with TsFresh

    Parameters
    ----------
    df_rolled : pandas DataFrame
        rolled DataFrame from roll_dataset
    feature : string, optional
        Extracted feature - only supports one at a time -
        and only features form tsfresh that dont need extra parameters.
        The default is "maximum".

    Returns
    -------
    extracted_features : pandas DataFrame
        Dataframe of extracted features

    """
    # extract features

    # fc_parameters = {"abs_energy", "maximum"}
    fc_parameters = {
        feature: None,
    }
    extracted_features = ts.extract_features(
        df_rolled,
        column_id="id",
        column_sort="time",
        default_fc_parameters=fc_parameters,
        n_jobs=0,
    )
    extracted_features.index = range(len(extracted_features))
    return extracted_features


def _select_relevant_windows(
    prep_roll,
    extracted_features,
    comparison_column_ex,
    fraction_max=0.25,
    window_size=1000,
    overlap=200,
    n_gridpoints=3,
    method="keep",
):
    """Writes n_gridpoints NaN's into the window_sizes with extracted features
    lower than fraction_max

    Parameters
    ----------
    prep_roll : pandas DataFrame
        input data - normally output from perpare_rolling(df)
    extracted_features : pandas Dataframe
        DataFrame of features
    comparison_column_ex: string - name of the extraced feature column
        it is build: comparison_column + '__' + feauture
    fraction_max : float
        percentage of the maximum of the extraced feature.
    window_size : int
        window size of the rolled segments  -the default is 1000.
    overlap : int, optional
         overlap between 2 adjecent windows -The default is 200.


    Returns
    -------
    df : pandas DataFrame relevant_windows
        dataframe with NaN's in the windows with too low extracted features

    """
    # get added up abs energy of interval x, if too low set None
    rolling_direction = window_size - overlap

    relevant_feature = extracted_features[comparison_column_ex]
    relevant_windows = prep_roll.copy()
    just_added_NaNs = False
    liste = []
    for i in range(len(extracted_features)):
        if relevant_feature[i] <= relevant_feature.max() * fraction_max:
            if just_added_NaNs is True:
                liste.append(
                    list(
                        range(
                            0 + i * rolling_direction,
                            window_size + i * rolling_direction,
                        )
                    )
                )

            else:
                liste.append(
                    list(
                        range(
                            0 + i * rolling_direction,
                            window_size + i * rolling_direction - n_gridpoints,
                        )
                    )
                )
                relevant_windows.iloc[
                    i * rolling_direction
                    + window_size
                    - n_gridpoints : i * rolling_direction
                    + window_size,
                    0 : relevant_windows.shape[1] - 2,
                ] = None
                just_added_NaNs = True
        else:
            just_added_NaNs = False

    index_liste = []
    """
    tail = (len(prep_roll)-window_size) % rolling_direction+1
    for i in range(tail):
        liste.append(len(prep_roll)-i-1)
    """
    liste = list(pd.core.common.flatten(liste))
    liste = list(set(liste))
    for i in range(len(liste)):
        index_liste.append(relevant_windows.index[liste[i]])
    if method == "keep":
        relevant_windows = relevant_windows.drop(index_liste, axis=0)
    elif method == "remove":
        relevant_windows = relevant_windows.loc[index_liste]
    return relevant_windows


def _polyfit_gridpoints(grid_points, prep_roll, order=3, verbose=False, n_gridpoints=3):
    """Fills gridpoints with polynomial regression

    Parameters
    ----------
    gridpoints : pandas DataFrame
        DataFrame with NaN's as gridpoints
    prep_roll : pandas DataFrame used to create time axis.
        DataFrame used to create time axis.
    order : int, optional
        Order of polynom The default is 3.
    verbose : boolean, optional
        If true plots polyfits. The default is False.
    n_gridpoints : TYPE, optional
        Number of gridpoints. The default is 3.

    Returns
    -------
    df : pandas DataFrame
        DataFrame with polynomial values at the gridpoints.
    """

    # add a null row at the start and reset time index
    delta_t = prep_roll.index[1] - prep_roll.index[0]
    line = pd.DataFrame(grid_points.iloc[:1], index=[-delta_t])
    grid_points = pd.concat([grid_points, line], ignore_index=False)
    grid_points.index = grid_points.index + delta_t
    poly_gridpoints = grid_points.sort_index()
    poly_gridpoints.iloc[0, :] = 0
    ts_time = prep_roll.iloc[: len(poly_gridpoints)]

    poly_gridpoints["time"] = ts_time.index.values
    poly_gridpoints.index = poly_gridpoints["time"]

    # %% smooth the gaps with polynomial values
    poly_gridpoints.interpolate(method="polynomial", order=order, inplace=True)

    return poly_gridpoints


def clean_timeseries(
    df,
    comparison_column,
    window_size=1000,
    overlap=800,
    feature="abs_energy",
    method="keep",
    n_gridpoints=3,
    percentage_max=0.05,
    order=3,
):
    """Removes segments of the data in which the extracted feature value is lower as
    percentage_max and fills the gaps with polynomial regression

    Parameters
    ----------
    df : input pandas DataFrame that shall be cleaned
    comparison_column: str, column that is used for the feature
        comparison with percentage max
    window_size : int, optional
        window size of the rolled segments - The default is 1000.
    overlap : int, optional
         overlap between 2 adjecent windows -The default is 200.
    feature : string, optional
        extracted feature - only supports one at a time -
        and only features form tsfresh that dont need extra parameters.
        The default is "maximum".
    method: string, optional
        * 'keep': keeps the windows which are extracted,
        * 'remove': removes the windows which are extracted
    n_gridpoints : TYPE, optional
        number of gridpoints. The default is 3.
    percentage_max : float, optional
        min percentage of the maximum to keep the window. The default is 0.05.
    order : int, optional
        order of polynom The default is 3.

    Returns
    -------
    df_poly : pandas DataFrame
        cleaned DataFrame

    """

    if not _HAVE_TSFRESH:
        raise ImportError(
            "tsfresh and dependencies are not installed. "
            "Use `pip install pylife[tsfresh]` to install it."
        )

    df_prep = _prepare_rolling(df)
    ts_time = df_prep.copy()
    # adding a row
    delta_t = ts_time.index[1] - ts_time.index[0]
    line = pd.DataFrame(ts_time.iloc[:1], index=[-delta_t])
    ts_time = pd.concat([ts_time, line], ignore_index=False)
    ts_time.index = ts_time.index + delta_t
    ts_time = ts_time.sort_index()

    ts_time["time"] = ts_time.index.values

    comparison_column_ex = comparison_column + "__" + feature
    df_rolled = _roll_dataset(df_prep, window_size=window_size, overlap=overlap)
    extracted_features = _extract_feature_df(df_rolled, feature)
    grid_points = _select_relevant_windows(
        df_prep,
        extracted_features,
        comparison_column_ex,
        percentage_max,
        window_size,
        overlap,
        method=method,
    )

    poly_gridpoints = _polyfit_gridpoints(
        grid_points, ts_time, order=order, verbose=False, n_gridpoints=n_gridpoints
    )

    # Remove NaN's at the end - should be maximum 2n
    cleaned = poly_gridpoints.dropna(axis=0, how="any")
    cleaned.pop("id")

    return cleaned



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pandas as pd
from pylife.stress import frequencysignal as fsig


def test_rms_psd():
    psd_nCode = pd.DataFrame([[400, 1.265, 1.003, 0.2959],
                              [800, 2.215, 2.291, 0.7709], [1200, 1.407, 1.479, 0.3021],
                              [1600, 1.083, 0.6476, 0.1402], [2000, 0.3998, 0.1892, 0.139],
                              [2400, 0.06213, 0.03343, 0.02849], [
                                  2800, 0.01297, 0.006958, 0.006106],
                              [3200, 0.006231, 0.005423, 0.004105], [
                                  3600, 0.008619, 0.006453, 0.005322],
                              [4000, 0.0119, 0.005596, 0.005025], [
                                  4400, 0.01317, 0.004346, 0.003961],
                              [4800, 0.01143, 0.001543, 0.002799], [
                                  5200, 0.007748, 5.323E-4, 0.001033],
                              [5600, 0.005334, 2.958E-4, 4.617E-4], [6000, 0.004206, 2.962E-4, 8.804E-4],
                              [6400, 0.002729, 2.179E-4, 0.001231]])
    psd_nCode = psd_nCode.set_index(psd_nCode.columns[0])
    rms = fsig.psdSignal.rms_psd(psd_nCode)
    rms.name = 'RMS'
    rms_nCode = pd.Series([47.72598934, 44.4876549, 24.20755918], name='RMS', index=[1, 2, 3])
    pd.testing.assert_series_equal(rms, rms_nCode, rtol=1)


def test_psd_smoother():
    psd_df = pd.DataFrame(data=[4, 4, 2, 2, 1, 1],
                          index=[1, 10, 10.01, 100, 100.1, 1000])
    f_fine = np.logspace(np.log10(psd_df.index.values.min()), np.log10(
        psd_df.index.values.max()), 1024)
    # Only 1 freq selected:
    test_psd = fsig.psdSignal.psd_smoother(psd_df, [50], 0)
    np.testing.assert_almost_equal(max(test_psd.values), 2)
    test_psd = fsig.psdSignal.psd_smoother(psd_df, [50], 1)

    pd.testing.assert_series_equal(fsig.psdSignal.rms_psd(test_psd),
                                   fsig.psdSignal.rms_psd(pd.DataFrame(10**np.interp(f_fine,
                                                                                     psd_df.index.values,
                                                                                     np.log10(psd_df.values.flatten())),
                                                                       index=f_fine)),
                                   rtol=1)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import unittest.mock as mock

from testbook import testbook

from IPython.core.profiledir import ProfileDir

pytestmark = pytest.mark.demos


@pytest.fixture(autouse=True, scope='session')
def ipython_profile(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp('ipython-dir-').as_posix()
    ProfileDir.create_profile_dir(tmp_path)
    with mock.patch.dict(os.environ, {'IPYTHONDIR': tmp_path}):
        yield


@pytest.fixture(autouse=True)
def change_workingdir_dir(monkeypatch):
    monkeypatch.chdir('demos')


@testbook('demos/hotspot_beam.ipynb')
def test_hotspot_beam(tb):
    with tb.patch('pyvista.Plotter'):
        tb.execute()

    tb.inject(
        """
        first_hotspot = mesh[mesh['hotspot'] == 1]
        assert len(first_hotspot) == 1296
        """
    )
    tb.inject(
        """
        second_hotspot = mesh[mesh['hotspot'] == 2]
        assert len(second_hotspot) == 1504
        """
    )
    tb.inject(
        """
        third_hotspot = mesh[mesh['hotspot'] == 3]
        assert len(third_hotspot) == 160
        """
    )


@testbook('demos/local_stress_with_FE.ipynb')
def test_local_stress_with_fe(tb):
    with tb.patch('pyvista.Plotter'):
        tb.execute()

    tb.inject("np.testing.assert_approx_equal(damage.max(), 0.0023, significant=2)")



@testbook('demos/psd_optimizer.ipynb', execute=True)
def test_psd_optimizer(tb):
    tb.inject(
        """
        np.testing.assert_allclose(
            [rms_psd(psdSignal.psd_smoother(psd, fsel, fit)) for fit in np.linspace(0, 1, 4)],
            [19.30855280765641, 11.13723910736734, 9.997749625841077, np.nan], rtol=1e-05
        )
        """
    )


@testbook('demos/ramberg_osgood.ipynb', execute=True)
def test_ramberg_osgood(tb):
    tb.inject(
        """
        assert monotone_strain.max() == hyst_strain.max()
        assert hyst_strain.max() == - hyst_strain.min()
        """
    )


@testbook('demos/time_series_handling.ipynb', execute=True, timeout=600)
def test_time_series_handling(tb):
    tb.inject(
        """
        assert 'envelope' in df_psd.columns
        """
    )


@testbook('demos/lifetime_calc.ipynb')
def test_lifetime_calc(tb):
    with tb.patch('pyvista.Plotter'):
        tb.execute()

    tb.inject(
        """
        np.testing.assert_approx_equal(fp_component, 3.04e-04, significant=2)
        %store -d rf_dict
        """
    )



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Until version 2.1.0 pyLife had a module for Bayesian Wöhler analysis.  The
Bayesian Wöhler analysis is not well established in the community, so it was a
bit ahead of its time.  Now that we have collected experiences with it, it
turns out, that its result are too often inaccurate.  Therefore we decided to
disable it.

That means, that the code remains in the code base, we are only preventing you
from importing it.  That means, if you want to experiment with the code, you
can grab the module and delete the exception raising.  We do not recommend to
use it for production, though.

If you want to propose an improvement of the code that leads to better results,
we are open for PRs (see :doc:`/CONTRIBUTING`) and might eventually enable it
again.
"""

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

raise NotImplementedError(
    "pyLife's Bayesian Wöhler analyzer has been shutdown. "
    "See documentation for details."
)

import sys
import numpy as np
import pandas as pd
import pytensor.tensor as pt
import pymc as pm
import bambi

from .elementary import Elementary
from .likelihood import Likelihood


class Bayesian(Elementary):
    """A Wöhler analyzer using Bayesian optimization

    Warning
    -------

    We are considering switching from pymc to GPyOpt as calculation engine in the
    future.  Maybe this will lead to breaking changes without new major release.
    """


    class _LogLike(pt.Op):
        """
        Specify what type of object will be passed and returned to the Op when it is
        called. In our case we will be passing it a vector of values (the parameters
        that define our model) and returning a single "scalar" value (the
        log-likelihood)

        http://mattpitkin.github.io/samplers-demo/pages/pymc-blackbox-likelihood/
        """
        itypes = [pt.dvector]  # expects a vector of parameter values when called
        otypes = [pt.dscalar]  # outputs a single scalar value (the log likelihood)

        def __init__(self, likelihood):
            """
            Initialise the Op with various things that our log-likelihood function
            requires. Below are the things that are needed in this particular
            example.

            Parameters
            ----------
            loglike:
                The log-likelihood function we've defined
            data:
                The "observed" data that our log-likelihood function takes in
            x:
                The dependent variable (aka 'load_cycle_limit') that our model requires

            """
            # add inputs as class attributes
            self.likelihood = likelihood

        def perform(self, node, inputs, outputs):
            # the method that is used when calling the Op
            var, = inputs  # this will contain my variables
            # mali_sum_lolli(var, self.zone_inf, self.load_cycle_limit):
            # call the log-likelihood function
            logl = self.likelihood.likelihood_infinite(var[0], var[1])

            outputs[0][0] = np.array(logl)  # output the log-likelihood

    def _specific_analysis(self, wc, nsamples=1000, **kw):
        self._nsamples = nsamples
        nburn = self._nsamples // 10

        tune = kw.pop('tune', 1000)
        random_seed = kw.pop('random_seed', None)

        SD_TS_trace = self._SD_TS_trace(tune=tune, random_seed=random_seed, **kw)
        slope_trace = self._slope_trace(tune=tune, random_seed=random_seed, **kw)
        slope = slope_trace.get('x').values[-1, nburn:].mean()
        intercept = slope_trace.get('Intercept').values[-1, nburn:].mean()

        SD = SD_TS_trace.get('SD').values[-1, nburn:].mean()
        ND = np.power(10., np.log10(SD) * slope + intercept)

        TN_trace = self._TN_trace(tune=tune, random_seed=random_seed, **kw)

        return pd.Series({
            'SD': SD,
            'TS': SD_TS_trace.get('TS').values[-1, nburn:].mean(),
            'ND': ND,
            'k_1': -slope,
            'TN': TN_trace.get('mu').values[-1, nburn:].mean(),
        })

    def _slope_trace(self, chains=2, random_seed=None, tune=1000, **kw):
        data_dict = pd.DataFrame({
            'x': np.log10(self._fd.fractures.load),
            'y': np.log10(self._fd.fractures.cycles.to_numpy())
        })
        with pm.Model():
            model = bambi.Model('y ~ x', data_dict, family='t')
            fitted = model.fit(
                draws=self._nsamples,
                cores=self._core_num(),
                tune=tune,
                chains=chains,
                random_seed=random_seed,
                **kw
            )
            return fitted.posterior

    def _TN_trace(self, chains=3, random_seed=None, tune=1000, **kw):
        with pm.Model():
            log_N_shift = np.log10(self._pearl_chain_estimator.normed_cycles)
            stdev = pm.HalfNormal('stdev', sigma=1.3)  # sigma standard wert (log-normal/ beat Verteilung/exp lambda)
            mu = pm.Normal('mu', mu=log_N_shift.mean(), sigma=log_N_shift.std())
            _ = pm.Normal('y', mu=mu, sigma=stdev, observed=log_N_shift)

            trace_TN = pm.sample(self._nsamples,
                                 cores=self._core_num(),
                                 target_accept=0.99,
                                 random_seed=random_seed,
                                 chains=chains,
                                 tune=tune,
                                 **kw)

        return trace_TN.posterior

    def _SD_TS_trace(self, chains=3, random_seed=None, tune=1000, **kw):
        loglike = self._LogLike(Likelihood(self._fd))

        with pm.Model():
            inf_load = self._fd.infinite_zone.load
            SD = pm.Normal('SD', mu=inf_load.mean(), sigma=inf_load.std()*5)
            TS_inv = pm.Lognormal('TS', mu=np.log10(1.1), sigma=0.3)

            # convert m and c to a tensor vector
            print(type(SD))
            print(type(TS_inv))
            var = pt.as_tensor_variable([SD, TS_inv])

            pm.Potential('likelihood', loglike(var))

            trace_SD_TS = pm.sample(self._nsamples,
                                    cores=self._core_num(),
                                    chains=chains,
                                    random_seed=random_seed,
                                    discard_tuned_samples=True,
                                    tune=tune,
                                    **kw)

        return trace_SD_TS.posterior

    def _core_num(self):
        return 1 if sys.platform.startswith('win') else None



# Copyright (c) 2019-2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

import pylife.utils.functions as functions
from pylife.utils.probability_data import ProbabilityFit


class PearlChainProbability(ProbabilityFit):
    """Shift all the data point to a normalized load level.

    Pearl chain method: consists of shifting the fractured data to a median
    load level.  The shifted data points are assigned to a Rossow failure
    probability.  The scatter in load-cycle direction can be computed from the
    probability net.

    Parameters
    ----------
    fracutres: pd.DataFrame consisting `load` and `cycles`
        The data point of the fractures to be shifted.

    slope: float
        The ``k_1`` slope the data is to be shifted along.

    """

    def __init__(self, fractures, slope):
        self._normed_load = fractures.load.mean()
        self._normed_cycles = np.sort(fractures.cycles * ((self._normed_load/fractures.load)**(slope)))

        fp = functions.rossow_cumfreqs(len(self._normed_cycles))
        super().__init__(fp, self._normed_cycles)

    @property
    def normed_load(self):
        """The normalized (shifted) load level."""
        return self._normed_load

    @property
    def normed_cycles(self):
        """The cycles shifted to the normalized load level."""
        return self._normed_cycles



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import numpy as np
import pandas as pd

from hypothesis import given, settings, strategies as st

from pylife.materiallaws import WoehlerCurve


@pytest.fixture
def wc_data():
    return pd.Series({
    'k_1': 7.,
    'TN': 1.75,
    'ND': 1e6,
    'SD': 300.0
})


def test_woehler_accessor(wc_data):
    wc = wc_data.drop('TN')

    for key in wc.index:
        wc_miss = wc.drop(key)
        with pytest.raises(AttributeError):
            wc_miss.woehler


def test_woehler_transform_probability():
    wc_50 = pd.Series({
        'k_1': 2,
        'k_2': np.inf,
        'TS': 2.,
        'TN': 9.,
        'ND': 3e6,
        'SD': 300 * np.sqrt(2.),
        'failure_probability': 0.5
    }).sort_index()

    transformed_90 = wc_50.woehler.transform_to_failure_probability(0.9).to_pandas()
    pd.testing.assert_series_equal(transformed_90[['SD', 'ND', 'failure_probability']],
                                   pd.Series({'SD': 600.0, 'ND': 4.5e6, 'failure_probability': 0.9}))
    transformed_back = transformed_90.woehler.transform_to_failure_probability(0.5).to_pandas()
    pd.testing.assert_series_equal(transformed_back, wc_50)

    transformed_10 = wc_50.woehler.transform_to_failure_probability(0.1).to_pandas()
    pd.testing.assert_series_equal(transformed_10[['SD', 'ND', 'failure_probability']],
                                   pd.Series({'SD': 300.0, 'ND': 2e6, 'failure_probability': 0.1}))
    transformed_back = transformed_10.woehler.transform_to_failure_probability(0.5).to_pandas()
    pd.testing.assert_series_equal(transformed_back, wc_50)


def test_woehler_transform_probability_multiple():
    wc_50 = pd.Series({
        'k_1': 2,
        'k_2': np.inf,
        'TS': 2.,
        'TN': 9.,
        'ND': 3e6,
        'SD': 300 * np.sqrt(2.),
        'failure_probability': 0.5
    }).sort_index()

    transformed = wc_50.woehler.transform_to_failure_probability([.1, .9]).to_pandas()
    expected = pd.DataFrame({
        'k_1': [2., 2.],
        'k_2': [np.inf, np.inf],
        'TS': [2., 2.],
        'TN': [9., 9.],
        'ND': [2e6, 4.5e6],
        'SD': [300., 600.],
        'failure_probability': [0.1, 0.9]
    })

    pd.testing.assert_frame_equal(transformed, expected, check_like=True)

    transformed_back = transformed.woehler.transform_to_failure_probability([0.5, 0.5]).to_pandas()
    expected = pd.DataFrame({
        'k_1': [2., 2.],
        'k_2': [np.inf, np.inf],
        'TS': [2., 2.],
        'TN': [9., 9.],
        'ND': [3e6, 3e6],
        'SD': [300. * np.sqrt(2.), 300. * np.sqrt(2.)],
        'failure_probability': [0.5, 0.5]
    })

    pd.testing.assert_frame_equal(transformed_back, expected, check_like=True)


def test_woehler_transform_probability_SD_0():
    wc_50 = pd.Series({
        'k_1': 2,
        'k_2': np.inf,
        'TS': 2.,
        'TN': 9.,
        'ND': 3e6,
        'SD': 0.0,
        'failure_probability': 0.5
    }).sort_index()

    transformed_90 = wc_50.woehler.transform_to_failure_probability(0.9).to_pandas()
    pd.testing.assert_series_equal(transformed_90[['SD', 'ND', 'failure_probability']],
                                   pd.Series({'SD': 0, 'ND': 9e6, 'failure_probability': 0.9}))
    transformed_back = transformed_90.woehler.transform_to_failure_probability(0.5).to_pandas()
    pd.testing.assert_series_equal(transformed_back, wc_50)

    transformed_10 = wc_50.woehler.transform_to_failure_probability(0.1).to_pandas()
    pd.testing.assert_series_equal(transformed_10[['SD', 'ND', 'failure_probability']],
                                   pd.Series({'SD': 0.0, 'ND': 1e6, 'failure_probability': 0.1}))
    transformed_back = transformed_10.woehler.transform_to_failure_probability(0.5).to_pandas()
    pd.testing.assert_series_equal(transformed_back, wc_50)


def test_woehler_basquin_cycles_50_single_load_single_wc(wc_data):
    load = 500.

    cycles = wc_data.woehler.basquin_cycles(load)
    expected_cycles = 27994

    np.testing.assert_allclose(cycles, expected_cycles, rtol=1e-4)


def test_woehler_basquin_cycles_50_multiple_load_single_wc(wc_data):
    load = [200., 300., 400., 500.]

    cycles = wc_data.woehler.basquin_cycles(load)
    expected_cycles = [np.inf, 1e6,  133484,    27994]

    np.testing.assert_allclose(cycles, expected_cycles, rtol=1e-4)


def test_woehler_basquin_cycles_50_single_load_multiple_wc():
    load = 400.

    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6]
    })

    cycles = wc.woehler.basquin_cycles(load)
    expected_cycles = [7.5e5, 1e6, np.inf]

    np.testing.assert_allclose(cycles, expected_cycles, rtol=1e-4)


def test_woehler_basquin_cycles_50_multiple_load_multiple_wc():
    load = [3000., 400., 500.]

    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6]
    })

    cycles = wc.woehler.basquin_cycles(load)
    expected_cycles = [1e5, 1e6, 1e6]

    np.testing.assert_allclose(cycles, expected_cycles, rtol=1e-4)


def test_woehler_basquin_cycles_50_multiple_load_multiple_wc_aligned_index():
    index = pd.Index([1, 2, 3], name='element_id')
    load = pd.Series([3000., 400., 500.], index=index)

    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6]
    }, index=index)

    cycles = wc.woehler.basquin_cycles(load)
    expected_cycles = pd.Series([1e5, 1e6, 1e6], index=index)

    pd.testing.assert_series_equal(cycles, expected_cycles, rtol=1e-4)


def test_woehler_basquin_cycles_50_multiple_load_multiple_wc_cross_index():
    load = pd.Series([3000., 400., 500.], index=pd.Index([1, 2, 3], name='scenario'))

    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6]
    }, pd.Index([1, 2, 3], name='element_id'))

    cycles = wc.woehler.basquin_cycles(load)
    expected_index = pd.MultiIndex.from_tuples([
        (1, 1), (1, 2), (1, 3),
        (2, 1), (2, 2), (2, 3),
        (3, 1), (3, 2), (3, 3),
    ], names=['element_id', 'scenario'])

    pd.testing.assert_index_equal(cycles.index, expected_index)


def test_woehler_basquin_cycles_50_same_k(wc_data):
    load = [200., 300., 400., 500.]

    wc = wc_data.copy()
    wc['k_2'] = wc['k_1']
    cycles = wc.woehler.basquin_cycles(load)

    calculated_k = - (np.log(cycles[-1]) - np.log(cycles[0])) / (np.log(load[-1]) - np.log(load[0]))
    np.testing.assert_approx_equal(calculated_k, wc.k_1)


def test_woehler_basquin_cycles_10_90(wc_data):
    load = [200., 300., 400., 500.]

    cycles_10 = wc_data.woehler.basquin_cycles(load, 0.1)[1:]
    cycles_90 = wc_data.woehler.basquin_cycles(load, 0.9)[1:]

    expected = [np.inf, 1.75, 1.75]
    np.testing.assert_allclose(cycles_90/cycles_10, expected)


def test_woehler_basquin_load_50_single_cycles_single_wc(wc_data):
    cycles = 27994

    load = wc_data.woehler.basquin_load(cycles)
    expected_load = 500.

    np.testing.assert_allclose(load, expected_load, rtol=1e-4)


def test_woehler_basquin_load_50_multiple_cycles_single_wc(wc_data):
    cycles = [np.inf, 1e6,  133484,    27994]

    load = wc_data.woehler.basquin_load(cycles)
    expected_load = [300., 300., 400., 500.]

    np.testing.assert_allclose(load, expected_load, rtol=1e-4)


def test_woehler_basquin_load_single_cycles_multiple_wc():
    cycles = 1e6
    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e5, 1e6, 1e6]
    })

    load = wc.woehler.basquin_load(cycles)
    expected_load = [300., 400., 500.]
    np.testing.assert_allclose(load, expected_load)


def test_woehler_basquin_load_multiple_cycles_multiple_wc():
    cycles = [1e5, 1e6, 1e7]
    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6]
    })

    load = wc.woehler.basquin_load(cycles)
    expected_load = [3000., 400., 500.]
    np.testing.assert_allclose(load, expected_load)


def test_woehler_basquin_load_multiple_cycles_multiple_wc_aligned_index():
    index = pd.Index([1, 2, 3], name='element_id')
    cycles = pd.Series([1e5, 1e6, 1e7], index=index)

    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6],
    }, index=index)

    expected_load = pd.Series([3000., 400., 500.], index=cycles.index)
    load = wc.woehler.basquin_load(cycles)

    pd.testing.assert_series_equal(load, expected_load)


def test_woehler_basquin_load_multiple_cycles_multiple_wc_cross_index():
    cycles = pd.Series([1e5, 1e6, 1e7], index=pd.Index([1, 2, 3], name='scenario'))

    wc = pd.DataFrame({
        'k_1': [1., 2., 2.],
        'SD': [300., 400., 500.],
        'ND': [1e6, 1e6, 1e6],
    }, index=pd.Index([1, 2, 3], name='element_id'))

    expected_index = pd.MultiIndex.from_tuples([
        (1, 1), (1, 2), (1, 3),
        (2, 1), (2, 2), (2, 3),
        (3, 1), (3, 2), (3, 3),
    ], names=['element_id', 'scenario'])

    load = wc.woehler.basquin_load(cycles)

    pd.testing.assert_index_equal(load.index, expected_index)


def test_woehler_basquin_load_50_same_k(wc_data):
    cycles = [1e7, 1e6, 1e5, 1e4]

    wc = wc_data.copy()
    wc['k_2'] = wc['k_1']

    load = wc.woehler.basquin_load(cycles)
    calculated_k = - (np.log(cycles[-1]) - np.log(cycles[0])) / (np.log(load[-1]) - np.log(load[0]))
    np.testing.assert_approx_equal(calculated_k, wc.k_1)


def test_woehler_basquin_load_10_90(wc_data):
    cycles = [1e2, 1e7]

    load_10 = wc_data.woehler.basquin_load(cycles, 0.1)
    load_90 = wc_data.woehler.basquin_load(cycles, 0.9)

    expected = np.full_like(cycles, 1.75 ** (1./7.))

    np.testing.assert_allclose(load_90/load_10, expected, rtol=1e-4)


def test_woehler_basquin_load_integer_cycles(wc_data):
    wc_data.woehler.basquin_load(1000)


def test_woehler_basquin_cycles_integer_load(wc_data):
    wc_data.woehler.basquin_cycles(200)


@pytest.fixture
def wc_int_overflow():
    return pd.Series({
        'k_1': 4.,
        'SD': 100.,
        'ND': 1e6
    })


def test_woehler_integer_overflow_scalar(wc_int_overflow):
    assert wc_int_overflow.woehler.basquin_cycles(50) > 0.0


def test_woehler_integer_overflow_list(wc_int_overflow):
    assert (wc_int_overflow.woehler.basquin_cycles([50, 50]) > 0.0).all()


def test_woehler_integer_overflow_series(wc_int_overflow):
    load = pd.Series([50, 50], index=pd.Index(['foo', 'bar']))
    cycles = wc_int_overflow.woehler.basquin_cycles(load)
    assert (cycles > 0.0).all()
    pd.testing.assert_index_equal(cycles.index, load.index)


def test_woehler_ND(wc_data):
    assert wc_data.woehler.ND == 1e6


def test_woehler_SD(wc_data):
    assert wc_data.woehler.SD == 300


def test_woehler_k_1(wc_data):
    assert wc_data.woehler.k_1 == 7.


def test_woehler_TS_and_TN_guessed():
    wc = pd.Series({
        'k_1': 0.5,
        'SD': 300,
        'ND': 1e6
    })
    assert wc.woehler.TN == 1.0
    assert wc.woehler.TS == 1.0


def test_woehler_TS_guessed(wc_data):
    wc = wc_data.copy()
    wc['k_1'] = 0.5
    wc['TN'] = 1.5

    assert wc.woehler.TS == (1.5 * 1.5)


def test_woehler_TN_guessed():
    wc = pd.Series({
        'k_1': 0.5,
        'SD': 300,
        'ND': 1e6,
        'TS': 1.5 * 1.5
    })

    assert wc.woehler.TN == 1.5


def test_woehler_TS_given(wc_data):
    wc_full = wc_data.copy()
    wc_full['TS'] = 1.25
    assert wc_full.woehler.TS == 1.25


def test_woehler_TN_given(wc_data):
    wc_full = wc_data.copy()
    wc_full['TN'] = 1.75
    assert wc_full.woehler.TN == 1.75


def test_woehler_pf_guessed(wc_data):
    assert wc_data.woehler.failure_probability == 0.5


def test_woehler_pf_given(wc_data):
    wc = wc_data.copy()
    wc['failure_probability'] = 0.1
    assert wc.woehler.failure_probability == 0.1


def test_woehler_miner_original_as_default(wc_data):
    assert wc_data.woehler.k_2 == np.inf


def test_woehler_miner_original_as_request(wc_data):
    wc_data['k_2'] = wc_data.k_1
    assert wc_data.woehler.miner_original().k_2 == np.inf


def test_woehler_miner_original_new_object(wc_data):
    orig = wc_data.woehler
    assert orig.miner_original() is not orig


def test_woehler_miner_elementary(wc_data):
    assert wc_data.woehler.miner_elementary().k_2 == wc_data.k_1
    assert wc_data.woehler.miner_elementary().to_pandas().k_2 == wc_data.k_1


def test_woehler_miner_elementary_new_object(wc_data):
    orig = wc_data.woehler
    assert orig.miner_elementary() is not orig


def test_woehler_miner_haibach(wc_data):
    assert wc_data.woehler.miner_haibach().k_2 == 13.0
    assert wc_data.woehler.miner_haibach().to_pandas().k_2 == 13.0


def test_woehler_miner_haibach_new_object(wc_data):
    orig = wc_data.woehler
    assert orig.miner_haibach() is not orig


def test_woehler_to_pandas(wc_data):
    expected = pd.Series({
        'k_1': 0.5,
        'k_2': np.inf,
        'TN': 1.75,
        'TS': 1.75 * 1.75,
        'ND': 1e6,
        'SD': 300.0,
        'failure_probability': 0.5,
    }).sort_index()
    wc = wc_data.copy()
    wc['k_1'] = 0.5

    pd.testing.assert_series_equal(wc.woehler.to_pandas().sort_index(), expected)

    wc = wc_data.copy()
    wc['k_1'] = 0.5
    del wc['TN']

    expected['TN'] = 1.
    expected['TS'] = 1.

    pd.testing.assert_series_equal(wc.woehler.to_pandas().sort_index(), expected)


@pytest.mark.parametrize('pf', [0.1, 0.5, 0.9])
def test_woehler_miner_original_homogenious_load(pf):
    cycles = np.logspace(3., 7., 50)
    wc = pd.Series({
        'k_1': 7.,
        'TN': 1.75,
        'ND': 1e6,
        'SD': 300.0
    })
    load = wc.woehler.basquin_load(cycles, failure_probability=pf)
    assert (np.diff(load) <= 0.).all()
    assert (np.diff(np.diff(load)) >= 0.).all()


@pytest.mark.parametrize('pf', [0.1, 0.5, 0.9])
def test_woehler_miner_original_homogenious_cycles(pf, wc_data):
    load = np.logspace(3., 2., 50)
    cycles = wc_data.woehler.basquin_cycles(load, failure_probability=pf)
    assert (np.diff(cycles[np.isfinite(cycles)]) > 0.).all()


def test_broadcast_load_cycles_clashing_index():
    """Don't know exactly what's going on there."""
    wc = pd.DataFrame({
        'SD': [500., 500., 300.],
        'k_1': [6., 6., 6.],
        'TS': [2., 1.0000001, 1.25],
        'ND': [1e6, 1e6, 1e6]
    })

    cycles = pd.Series([1e6, 1e6, 1e6])

    result = wc.woehler.basquin_load(cycles)
    expected = pd.Series(
        [500., 500., 500., 500., 500., 500., 300., 300., 300],
        index=pd.MultiIndex.from_product([[0, 1, 2], [0, 1, 2]])
    )

    pd.testing.assert_series_equal(result, expected)


@settings(deadline=None)
@given(st.floats(min_value=10., max_value=500.),
       st.floats(min_value=1.0, max_value=10.0),
       st.floats(min_value=1e2, max_value=1e7),
       st.floats(min_value=1.0, max_value=1e9))
def test_load_is_basquin_load(SD, k_1, ND, cycles):
    wc = WoehlerCurve.from_parameters(SD=SD, k_1=k_1, ND=ND)
    assert wc.load(cycles) == wc.basquin_load(cycles)


@settings(deadline=None)
@given(st.floats(min_value=10., max_value=500.),
       st.floats(min_value=1.0, max_value=10.0),
       st.floats(min_value=1e2, max_value=1e7),
       st.floats(min_value=1.0, max_value=1000.0))
def test_cycles_is_basquin_cycles(SD, k_1, ND, load):
    wc = WoehlerCurve.from_parameters(SD=SD, k_1=k_1, ND=ND)
    assert wc.cycles(load) == wc.basquin_cycles(load)



# Copyright (c) 2020-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from abc import ABC, abstractmethod


class VMAPDataset(ABC):
    """
    Base class of VMAP dataset types
    """
    def __init__(self, identifier):
        self._identifier = identifier

    def set_identifier(self, identifier):
        self._identifier = identifier

    @property
    @abstractmethod
    def attributes(self):
        pass

    @property
    @abstractmethod
    def dtype(self):
        pass

    @property
    @abstractmethod
    def dataset_name(self):
        pass

    @property
    def group_path(self):
        return '/VMAP/SYSTEM'

    @property
    def compound_dataset(self):
        return True



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import pandas as pd
import numpy as np

import pylife.stress.collective


@pytest.fixture
def lc_matrix_range_mean():
    range_intervals = pd.interval_range(0., 12., 3)
    mean_intervals = pd.interval_range(-2., 4., 3)
    index = pd.MultiIndex.from_product([range_intervals, mean_intervals], names=['range', 'mean'])
    return pd.Series(1, index=index)


@pytest.fixture
def lc_matrix_from_to():
    from_intervals = pd.interval_range(-4., 8., 3)
    to_intervals = pd.interval_range(-2., 4., 3)
    index = pd.MultiIndex.from_product([from_intervals, to_intervals], names=['from', 'to'])
    return pd.Series(1, index=index)


def test_lc_matrix_fail_empty_fail():
    lcm = pd.Series(dtype=np.float64)
    with pytest.raises(AttributeError, match=r"Load collective matrix needs .* index levels"):
        lcm.load_collective


def test_lc_matrix_fail_no_interval_index():
    lcm = pd.Series([], index=pd.MultiIndex.from_tuples([], names=['from', 'to']), dtype=np.float64)
    with pytest.raises(AttributeError, match="Index of a load collective matrix must be pandas.IntervalIndex."):
        lcm.load_collective


def test_lc_matrix_fail_from_to_only_one_interval_index():
    emtpy_interval_index = pd.interval_range(0., 0., 0)
    index = pd.MultiIndex.from_product([emtpy_interval_index, []], names=['from', 'to'])
    lcm = pd.Series([], index=index, dtype=np.float64)
    with pytest.raises(AttributeError, match="Index of a load collective matrix must be pandas.IntervalIndex."):
        lcm.load_collective

    index = pd.MultiIndex.from_product([emtpy_interval_index, []], names=['to', 'from'])
    lcm = pd.Series([], index=index, dtype=np.float64)
    with pytest.raises(AttributeError, match="Index of a load collective matrix must be pandas.IntervalIndex."):
        lcm.load_collective


def test_lc_matrix_fail_range_interval_index():
    index = pd.Index([], name='range')
    lcm = pd.Series([], index=index, dtype=np.float64)
    with pytest.raises(AttributeError, match="Index of a load collective matrix must be pandas.IntervalIndex."):
        lcm.load_collective


def test_lc_matrix_range_no_mean_interval_index():
    emtpy_interval_index = pd.interval_range(0., 0., 0)
    index = pd.IntervalIndex(emtpy_interval_index, name='range')
    lcm = pd.Series([], index=index, dtype=np.float64)
    lcm.load_collective


def test_lc_matrix_fail_range_mean_only_one_interval_index():
    emtpy_interval_index = pd.interval_range(0., 0., 0)
    index = pd.MultiIndex.from_product([emtpy_interval_index, []], names=['range', 'mean'])
    lcm = pd.Series([], index=index, dtype=np.float64)
    with pytest.raises(AttributeError, match="Index of a load collective matrix must be pandas.IntervalIndex."):
        lcm.load_collective

    index = pd.MultiIndex.from_product([emtpy_interval_index, []], names=['mean', 'range'])
    lcm = pd.Series([], index=index, dtype=np.float64)
    with pytest.raises(AttributeError, match="Index of a load collective matrix must be pandas.IntervalIndex."):
        lcm.load_collective


def test_lc_matrix_valid_from_to(lc_matrix_from_to):
    lc_matrix_from_to.load_collective


def test_lc_matrix_valid_only_from(lc_matrix_from_to):
    lc_matrix_from_to.index = lc_matrix_from_to.index.droplevel('to')
    with pytest.raises(AttributeError, match=r"Load collective matrix needs .* index levels"):
        lc_matrix_from_to.load_collective


def test_rainflow_from_to_amplitude_left(lc_matrix_from_to):
    expected = pd.Series([1., 2., 3., 1., 0., 1., 3., 2., 1.],
                         name='amplitude',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_left().amplitude, expected)


def test_rainflow_from_to_amplitude_mid(lc_matrix_from_to):
    expected = pd.Series([0.5, 1.5, 2.5, 1.5, 0.5, 0.5, 3.5, 2.5, 1.5],
                         name='amplitude',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.amplitude, expected)


def test_rainflow_from_to_amplitude_right(lc_matrix_from_to):
    expected = pd.Series([0., 1., 2., 2., 1., 0., 4., 3., 2.],
                         name='amplitude',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_right().amplitude, expected)


@pytest.mark.parametrize('cycles_value', [1, 2])
def test_rainflow_from_to_amplitude_histogram(lc_matrix_from_to, cycles_value):
    expected_index = pd.IntervalIndex.from_arrays(
        [0., 0., 2., 0., 0., 0., 4., 2., 0.],
        [4., 6., 8., 6., 4., 4., 10., 8., 6.],
        name='amplitude'
    )
    expected = pd.Series(cycles_value, index=expected_index, name='cycles')
    matrix = lc_matrix_from_to * cycles_value

    result = matrix.load_collective.amplitude_histogram

    pd.testing.assert_series_equal(result, expected)


def test_rainflow_from_to_meanstress_left(lc_matrix_from_to):
    expected = pd.Series([-3., -2., -1., -1., 0., 1., 1., 2., 3.],
                         name='meanstress',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_left().meanstress, expected)


def test_rainflow_from_to_meanstress_mid(lc_matrix_from_to):
    expected = pd.Series([-1.5, -0.5, 0.5, 0.5, 1.5, 2.5, 2.5, 3.5, 4.5],
                         name='meanstress',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.meanstress, expected)


def test_rainflow_from_to_meanstress_right(lc_matrix_from_to):
    expected = pd.Series([0., 1., 2., 2., 3., 4., 4., 5., 6.],
                         name='meanstress',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_right().meanstress, expected)


def test_rainflow_from_to_upper_left(lc_matrix_from_to):
    expected = pd.Series([-2., 0., 2., 0., 0., 2., 4., 4., 4.],
                         name='upper',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_left().upper, expected)


def test_rainflow_from_to_upper_mid(lc_matrix_from_to):
    expected = pd.Series([-1., 1., 3., 2., 2., 3., 6., 6., 6.],
                         name='upper',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.upper, expected)


def test_rainflow_from_to_upper_right(lc_matrix_from_to):
    expected = pd.Series([0., 2., 4., 4., 4., 4., 8., 8., 8.],
                         name='upper',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_right().upper, expected)


def test_rainflow_from_to_lower_left(lc_matrix_from_to):
    expected = pd.Series([-4., -4., -4., -2., 0., 0., -2., 0., 2.],
                         name='lower',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_left().lower, expected)


def test_rainflow_from_to_lower_mid(lc_matrix_from_to):
    expected = pd.Series([-2., -2., -2., -1., 1., 2., -1., 1., 3.],
                         name='lower',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.lower, expected)


def test_rainflow_from_to_lower_right(lc_matrix_from_to):
    expected = pd.Series([0., 0., 0., 0., 2., 4., 0., 2., 4.],
                         name='lower',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_right().lower, expected)


def test_rainflow_R(lc_matrix_from_to):
    print(lc_matrix_from_to)
    expected = pd.Series([2., -np.inf, -2., -np.inf, 0., 0., -0.5, 0., 0.5],
                         name='R',
                         index=lc_matrix_from_to.index)
    pd.testing.assert_series_equal(lc_matrix_from_to.load_collective.use_class_left().R, expected)


def test_rainflow_cycles_from_to(lc_matrix_from_to):
    expected = pd.Series(1, index=lc_matrix_from_to.index, name='cycles')
    freq = lc_matrix_from_to.load_collective.cycles
    pd.testing.assert_series_equal(freq, expected)


def test_rainflow_from_to_scale_scalar(lc_matrix_from_to):
    from_intervals = pd.interval_range(-2., 4., 3)
    to_intervals = pd.interval_range(-1., 2., 3)
    expected_index = pd.MultiIndex.from_product([from_intervals, to_intervals], names=['from', 'to'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    scaled = lc_matrix_from_to.load_collective.scale(0.5)
    assert isinstance(scaled, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(scaled.to_pandas(), expected)


def test_rainflow_from_to_scale_series(lc_matrix_from_to):
    from_intervals = pd.IntervalIndex.from_arrays(
        [
            -2., -8., -2., -8., -2, -8.,
            0., 0., 0., 0., 0., 0.,
            2., 8., 2., 8., 2., 8.
        ], [
            0., 0., 0., 0., 0., 0.,
            2., 8., 2., 8., 2., 8.,
            4., 16., 4., 16., 4., 16.
        ]
    )
    to_intervals = pd.IntervalIndex.from_arrays(
        [
            -1., -4., 0., 0., 1., 4.,
            -1., -4., 0., 0., 1., 4.,
            -1., -4., 0., 0., 1., 4.
        ], [
            0., 0., 1., 4, 2., 8.,
            0., 0., 1., 4, 2., 8.,
            0., 0., 1., 4, 2., 8.
        ]
    )
    foo_index = pd.Index(['x', 'y']*9, name='foo')
    expected_index = pd.MultiIndex.from_arrays([from_intervals, to_intervals, foo_index], names=['from', 'to', 'foo'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    factors = pd.Series([0.5, 2.0], index=pd.Index(['x', 'y'], name='foo'), name='scale_factors')

    scaled = lc_matrix_from_to.load_collective.scale(factors)

    assert isinstance(scaled, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(scaled.to_pandas(), expected)


def test_rainflow_from_to_shift_scalar(lc_matrix_from_to):
    from_intervals = pd.interval_range(-0., 12., 3)
    to_intervals = pd.interval_range(2., 8., 3)
    expected_index = pd.MultiIndex.from_product([from_intervals, to_intervals], names=['from', 'to'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    shifted = lc_matrix_from_to.load_collective.shift(4.)
    assert isinstance(shifted, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(shifted.to_pandas(), expected)


def test_rainflow_from_to_shift_series(lc_matrix_from_to):
    from_intervals = pd.IntervalIndex.from_arrays(
        [
            -2., 0., -2., -0., -2., 0.,
            2., 4., 2., 4., 2., 4.,
            6., 8., 6., 8., 6., 8.
        ], [
            2., 4., 2., 4., 2., 4.,
            6., 8., 6., 8., 6., 8.,
            10., 12., 10., 12., 10., 12.
        ]
    )
    to_intervals = pd.IntervalIndex.from_arrays(
        [
            0., 2., 2., 4., 4., 6.,
            0., 2., 2., 4., 4., 6.,
            0., 2., 2., 4., 4., 6.
        ], [
            2., 4., 4., 6., 6., 8.,
            2., 4., 4., 6., 6., 8.,
            2., 4., 4., 6., 6., 8.
        ]
    )
    foo_index = pd.Index(['x', 'y']*9, name='foo')
    expected_index = pd.MultiIndex.from_arrays([from_intervals, to_intervals, foo_index], names=['from', 'to', 'foo'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    factors = pd.Series([2., 4.], index=pd.Index(['x', 'y'], name='foo'), name='shift_factors')

    shiftd = lc_matrix_from_to.load_collective.shift(factors)

    assert isinstance(shiftd, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(shiftd.to_pandas(), expected)


def test_lc_matrix_valid_range_mean(lc_matrix_range_mean):
    lc_matrix_range_mean.load_collective


def test_matrix_range_mean_amplitude_mid(lc_matrix_range_mean):
    expected = pd.Series([1., 1., 1., 3., 3., 3., 5., 5., 5.],
                         name='amplitude',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.amplitude, expected)


def test_matrix_range_mean_amplitude_left(lc_matrix_range_mean):
    expected = pd.Series([0., 0., 0., 2., 2., 2., 4., 4., 4.],
                         name='amplitude',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_left().amplitude, expected)


def test_matrix_range_mean_amplitude_right(lc_matrix_range_mean):
    expected = pd.Series([2., 2., 2., 4., 4., 4., 6., 6., 6.],
                         name='amplitude',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_right().amplitude, expected)


def test_matrix_rainge_mean_amplitude_histogram(lc_matrix_range_mean):
    expected_index = pd.IntervalIndex.from_arrays(
        [0., 0., 0., 2., 2., 2., 4., 4., 4.],
        [2., 2., 2., 4., 4., 4., 6., 6., 6.],
        name='amplitude'
    )
    result = lc_matrix_range_mean.load_collective.amplitude_histogram

    pd.testing.assert_index_equal(result.index, expected_index)


def test_matrix_range_mean_meanstress_left(lc_matrix_range_mean):
    expected = pd.Series([-2., 0., 2., -2., 0., 2., -2., 0., 2.],
                         name='meanstress',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_left().meanstress, expected)


def test_matrix_range_mean_meanstress_mid(lc_matrix_range_mean):
    expected = pd.Series([-1., 1., 3., -1., 1., 3., -1., 1., 3.],
                         name='meanstress',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.meanstress, expected)


def test_matrix_range_mean_meanstress_right(lc_matrix_range_mean):
    expected = pd.Series([0., 2., 4., 0., 2., 4., 0., 2., 4.],
                         name='meanstress',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_right().meanstress, expected)


def test_matrix_range_mean_meanstress_no_mean_defined():
    range_intervals = pd.interval_range(0., 12., 3)
    index = pd.IntervalIndex(range_intervals, name='range')
    rf = pd.Series(1., index=index)

    expected = pd.Series(np.zeros(3), name='meanstress', index=rf.index)
    pd.testing.assert_series_equal(rf.load_collective.use_class_right().meanstress, expected)


def test_matrix_range_mean_upper_left(lc_matrix_range_mean):
    expected = pd.Series([-2., 0., 2., 0., 2., 4., 2., 4., 6.],
                         name='upper',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_left().upper, expected)


def test_matrix_range_mean_upper_mid(lc_matrix_range_mean):
    expected = pd.Series([0., 2., 4., 2., 4., 6., 4., 6., 8.],
                         name='upper',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.upper, expected)


def test_matrix_range_mean_upper_right(lc_matrix_range_mean):
    expected = pd.Series([2., 4., 6., 4., 6., 8., 6., 8., 10.],
                         name='upper',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_right().upper, expected)


def test_matrix_range_mean_lower_left(lc_matrix_range_mean):
    expected = pd.Series([-2., 0., 2., -4., -2., 0., -6., -4., -2.],
                         name='lower',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_left().lower, expected)


def test_matrix_range_mean_lower_mid(lc_matrix_range_mean):
    expected = pd.Series([-2., 0., 2., -4., -2., 0., -6., -4., -2.],
                         name='lower',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.lower, expected)


def test_matrix_range_mean_lower_right(lc_matrix_range_mean):
    expected = pd.Series([-2., 0., 2., -4., -2., 0., -6., -4., -2.],
                         name='lower',
                         index=lc_matrix_range_mean.index)
    pd.testing.assert_series_equal(lc_matrix_range_mean.load_collective.use_class_right().lower, expected)


def test_matrix_range_mean_scale_scalar(lc_matrix_range_mean):
    range_intervals = pd.interval_range(0., 6., 3)
    mean_intervals = pd.interval_range(-1., 2., 3)
    expected_index = pd.MultiIndex.from_product([range_intervals, mean_intervals], names=['range', 'mean'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    scaled = lc_matrix_range_mean.load_collective.scale(0.5)
    assert isinstance(scaled, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(scaled.to_pandas(), expected)


def test_matrix_range_mean_foo_scale_scalar(lc_matrix_range_mean):
    foo_index = pd.Index([1, 2, 3], name='foo')
    levels = lc_matrix_range_mean.index.levels
    total_index = pd.MultiIndex.from_product([levels[0], levels[1], foo_index])
    lc_matrix = pd.Series(1, index=total_index)

    range_intervals = pd.interval_range(0., 6., 3)
    mean_intervals = pd.interval_range(-1., 2., 3)
    expected_index = pd.MultiIndex.from_product(
        [range_intervals, mean_intervals, foo_index],
        names=['range', 'mean', 'foo']
    )
    expected = pd.Series(1, index=expected_index, name='cycles')

    scaled = lc_matrix.load_collective.scale(0.5)
    assert isinstance(scaled, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(scaled.to_pandas(), expected)


def test_matrix_range_mean_scale_series(lc_matrix_range_mean):
    range_intervals = pd.IntervalIndex.from_arrays(
        [
            0., 0., 0., 0., 0., 0.,
            2., 8., 2., 8., 2., 8.,
            4., 16., 4., 16., 4., 16.
        ], [
            2., 8., 2., 8., 2., 8.,
            4., 16., 4., 16., 4., 16.,
            6., 24., 6., 24., 6., 24.
        ]
    )
    mean_intervals = pd.IntervalIndex.from_arrays(
        [
            -1., -4., 0., 0., 1., 4.,
            -1., -4., 0., 0., 1., 4.,
            -1., -4., 0., 0., 1., 4.
        ], [
            0., 0., 1., 4, 2., 8.,
            0., 0., 1., 4, 2., 8.,
            0., 0., 1., 4, 2., 8.
        ]
    )
    foo_index = pd.Index(['x', 'y']*9, name='foo')
    expected_index = pd.MultiIndex.from_arrays([range_intervals, mean_intervals, foo_index], names=['range', 'mean', 'foo'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    factors = pd.Series([0.5, 2.0], index=pd.Index(['x', 'y'], name='foo'), name='scale_factors')

    scaled = lc_matrix_range_mean.load_collective.scale(factors)

    assert isinstance(scaled, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(scaled.to_pandas(), expected)


def test_matrix_range_mean_shift_scalar(lc_matrix_range_mean):
    range_intervals = pd.interval_range(0., 12., 3)
    mean_intervals = pd.interval_range(2., 8., 3)
    expected_index = pd.MultiIndex.from_product([range_intervals, mean_intervals], names=['range', 'mean'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    shifted = lc_matrix_range_mean.load_collective.shift(4.)
    assert isinstance(shifted, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(shifted.to_pandas(), expected)


def test_matrix_range_mean_shift_series(lc_matrix_range_mean):
    range_intervals = pd.IntervalIndex.from_arrays(
        [
            0., 0., 0., 0., 0., 0.,
            4., 4., 4., 4., 4., 4.,
            8., 8., 8., 8., 8., 8.
        ], [
            4., 4., 4., 4., 4., 4.,
            8., 8., 8., 8., 8., 8.,
            12., 12., 12., 12., 12., 12.
        ]
    )
    mean_intervals = pd.IntervalIndex.from_arrays(
        [
            0., 2., 2., 4., 4., 6.,
            0., 2., 2., 4., 4., 6.,
            0., 2., 2., 4., 4., 6.,
        ], [
            2., 4., 4., 6., 6., 8.,
            2., 4., 4., 6., 6., 8.,
            2., 4., 4., 6., 6., 8.
        ]
    )
    foo_index = pd.Index(['x', 'y']*9, name='foo')
    expected_index = pd.MultiIndex.from_arrays([range_intervals, mean_intervals, foo_index], names=['range', 'mean', 'foo'])
    expected = pd.Series(1, index=expected_index, name='cycles')

    factors = pd.Series([2., 4.], index=pd.Index(['x', 'y'], name='foo'), name='shift_factors')

    shiftd = lc_matrix_range_mean.load_collective.shift(factors)

    assert isinstance(shiftd, pylife.stress.collective.LoadHistogram)
    pd.testing.assert_series_equal(shiftd.to_pandas(), expected)


@pytest.mark.parametrize('range_interval', [
    (pd.interval_range(0., 12., 3)),
    (pd.interval_range(0., 6., 3))
])
def test_matrix_cumulative_range_only_range(range_interval):
    idx = pd.IntervalIndex(range_interval, name='range')
    result = pd.Series(1, index=idx).load_collective.cumulated_range()

    expected = pd.Series([1, 1, 1], name='cumulated_cycles', index=idx)

    pd.testing.assert_series_equal(result, expected)


@pytest.mark.parametrize('range_interval', [
    (pd.interval_range(0., 12., 3)),
    (pd.interval_range(0., 6., 3))
])
def test_matrix_cumulative_range_range_mean(range_interval):
    range_idx = pd.IntervalIndex(range_interval, name='range')
    mean_idx = pd.IntervalIndex(pd.interval_range(0., 1., 3), name='mean')
    idx = pd.MultiIndex.from_product([range_idx, mean_idx])
    print(pd.Series(1, index=idx))
    result = pd.Series(1, index=idx).load_collective.cumulated_range()

    expected = pd.Series([1, 2, 3, 1, 2, 3, 1, 2, 3], name='cumulated_cycles', index=idx)

    pd.testing.assert_series_equal(result, expected)
